from pathlib import Path
from typing import cast

from openpyxl import Workbook
from openpyxl.cell import Cell
from openpyxl.styles import Alignment, Border, Font, NamedStyle, PatternFill, Side
from openpyxl.worksheet.worksheet import Worksheet

from cognite.neat.rules.models._base import EntityTypes

from ._base import BaseExporter


class ExcelExporter(BaseExporter[Workbook]):
    """Class for exporting transformation rules object to excel file."""

    def _export_to_file(self, filepath: Path) -> None:
        """Exports transformation rules to excel file."""
        data = self.export()
        try:
            data.save(filepath)
        finally:
            data.close()

    def export(self) -> Workbook:
        """Generates workbook from transformation rules."""
        data = Workbook()
        # Remove default sheet named "Sheet"
        data.remove(data["Sheet"])
        # Serialize the rules to the excel file

        # map rules metadata to excel
        metadata = self.rules.metadata
        metadata_sheet = data.create_sheet("Metadata")

        # add each metadata property to the sheet as a row
        metadata_sheet.append(["prefix", metadata.prefix])
        metadata_sheet.append(["namespace", metadata.namespace])
        metadata_sheet.append(["external_id", metadata.suffix])
        metadata_sheet.append(["title", metadata.name])
        metadata_sheet.append(["description", metadata.description])
        metadata_sheet.append(["version", metadata.version])
        metadata_sheet.append(
            [
                "creator",
                (
                    ",".join(metadata.creator if isinstance(metadata.creator, list) else [metadata.creator])
                    if metadata.creator
                    else ""
                ),
            ]
        )
        metadata_sheet.append(
            [
                "contributor",
                (
                    ",".join(metadata.contributor if isinstance(metadata.contributor, list) else [metadata.contributor])
                    if metadata.contributor
                    else ""
                ),
            ]
        )
        metadata_sheet.append(["created", metadata.created])
        metadata_sheet.append(["updated", metadata.updated])
        metadata_sheet.append(["rights", metadata.rights])

        # map classes to excel sheet named "Classes" and add each class as a row
        classes_sheet = data.create_sheet("Classes")

        classes_sheet.append(["Solution model", "", "", "Knowledge acquisition", "", "", ""])
        classes_sheet.merge_cells("A1:C1")
        classes_sheet.merge_cells("D1:G1")
        classes_sheet.append(
            ["Class", "Description", "Parent Class", "Source", "Source Entity Name", "Match Type", "Comment"]
        )  # A  ... # G

        for class_ in self.rules.classes.values():
            classes_sheet.append(
                [
                    class_.class_id,
                    class_.description,
                    (
                        ",".join([parent_class.versioned_id for parent_class in class_.parent_class])
                        if class_.parent_class
                        else None
                    ),
                    str(class_.source),
                    class_.source_entity_name,
                    class_.match_type,
                    class_.comment,
                ]
            )

        # map properties to excel sheet named "Properties" and add each property as a row
        properties_sheet = data.create_sheet("Properties")
        properties_sheet.append(
            [
                "Solution model",  # A
                "",  # B
                "",  # C
                "",  # D
                "",  # E
                "",  # F
                "Solution classic CDF resources",  # G
                "",  # H
                "",  # I
                "",  # J
                "",  # K
                "",  # L
                "Transformation rules",  # M
                "",  # N
                "Knowledge acquisition",  # O
                "",  # P
                "",  # Q
                "",  # R
            ]
        )
        properties_sheet.merge_cells("A1:F1")
        properties_sheet.merge_cells("G1:L1")
        properties_sheet.merge_cells("M1:N1")
        properties_sheet.merge_cells("O1:R1")
        properties_sheet.append(
            [
                "Class",  # A
                "Property",  # B
                "Description",  # C
                "Type",  # D
                "Min Count",  # E
                "Max Count",  # F
                "Resource Type",  # G
                "Resource Type Property",  # H
                "Relationship Source Type",  # I
                "Relationship Target Type",  # J
                "Relationship Label",  # K
                "Relationship ExternalID Rule",  # L
                "Rule Type",  # M
                "Rule",  # N
                "Source",  # O
                "Source Entity Name",  # P
                "Match Type",  # Q
                "Comment",  # R
            ]
        )

        for property_ in self.rules.properties.values():
            properties_sheet.append(
                [
                    property_.class_id,  # A
                    property_.property_id,  # B
                    property_.description,  # C
                    (
                        property_.expected_value_type.versioned_id
                        if property_.property_type == EntityTypes.object_property
                        else property_.expected_value_type.suffix
                    ),  # D
                    property_.min_count,  # E
                    property_.max_count,  # F
                    ",".join(property_.cdf_resource_type) if property_.cdf_resource_type else "",  # G
                    ",".join(property_.resource_type_property) if property_.resource_type_property else "",  # H
                    property_.source_type,  # I
                    property_.target_type,  # J
                    property_.label,  # K
                    property_.relationship_external_id_rule,  # L
                    property_.rule_type,  # M
                    property_.rule,  # N
                    str(property_.source),  # O
                    property_.source_entity_name,  # P
                    property_.match_type,  # Q
                    property_.comment,  # R
                ]
            )

        prefixes_sheet = data.create_sheet("Prefixes")
        prefixes_sheet.append(["Prefix", "URI"])  # A  # B

        for prefix, uri in self.rules.prefixes.items():
            prefixes_sheet.append([prefix, uri])  # A  # B

        return self._set_header_style(data)

    @staticmethod
    def _set_header_style(data: Workbook):
        """Sets the header style for all sheets in the self.workbook"""
        style = NamedStyle(name="header style")
        style.font = Font(bold=True, size=16)
        side = Side(style="thin", color="000000")
        style.border = Border(left=side, right=side, top=side, bottom=side)
        data.add_named_style(style)

        for sheet in data.sheetnames:
            if sheet == "Metadata":
                continue
            if sheet == "Classes" or sheet == "Properties":
                sheet_obj = cast(Worksheet, data[sheet])
                if sheet == "Classes":
                    sheet_obj.freeze_panes = "A3"
                else:
                    sheet_obj.freeze_panes = "D3"

                for cell in sheet_obj[1]:
                    cell = cast(Cell, cell)  # type: ignore[index]
                    cell.style = style
                    cell.fill = PatternFill("solid", start_color="D5B2CF")
                    cell.alignment = Alignment(horizontal="center", vertical="center")
                for cell in sheet_obj[2]:
                    cell = cast(Cell, cell)  # type: ignore[index]
                    cell.style = style
                    cell.fill = PatternFill("solid", start_color="D5DBD5")
                    cell.alignment = Alignment(horizontal="center", vertical="center")
                    adjusted_width = (len(str(cell.value)) + 5) * 1.2
                    cast(Worksheet, data[sheet]).column_dimensions[cell.column_letter].width = adjusted_width

        return data



from pathlib import Path
from typing import Literal, cast, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.containers import BTreeIndex, InvertedIndex
from cognite.client.utils import ms_to_datetime

from cognite.neat.rules import issues
from cognite.neat.rules.importers._base import BaseImporter, Rules
from cognite.neat.rules.issues import IssueList
from cognite.neat.rules.models._rules import DMSRules, DMSSchema, RoleTypes
from cognite.neat.rules.models._rules._types import (
    ClassEntity,
    ContainerEntity,
    DMSValueType,
    Undefined,
    Unknown,
    ViewEntity,
    ViewPropEntity,
)
from cognite.neat.rules.models._rules.base import ExtensionCategory, SchemaCompleteness
from cognite.neat.rules.models._rules.dms_architect_rules import (
    DMSContainer,
    DMSMetadata,
    DMSProperty,
    DMSView,
    SheetList,
)


class DMSImporter(BaseImporter):
    def __init__(self, schema: DMSSchema, metadata: DMSMetadata | None = None):
        self.schema = schema
        self.metadata = metadata

    @classmethod
    def from_data_model_id(cls, client: CogniteClient, data_model_id: DataModelIdentifier) -> "DMSImporter":
        """Create a DMSImporter ready to convert the given data model to rules.

        Args:
            client: Instantiated CogniteClient to retrieve data model.
            data_model_id: Data Model to retrieve.

        Returns:
            DMSImporter: DMSImporter instance
        """
        data_models = client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        if len(data_models) == 0:
            raise ValueError(f"Data model {data_model_id} not found")
        data_model = data_models.latest_version()
        schema = DMSSchema.from_data_model(client, data_model)
        description, creator = DMSMetadata._get_description_and_creator(data_model.description)

        created = ms_to_datetime(data_model.created_time)
        updated = ms_to_datetime(data_model.last_updated_time)

        metadata = DMSMetadata(
            schema_=SchemaCompleteness.complete,
            extension=ExtensionCategory.addition,
            space=data_model.space,
            external_id=data_model.external_id,
            name=data_model.name or data_model.external_id,
            version=data_model.version or "0.1.0",
            updated=updated,
            created=created,
            creator=creator,
            description=description,
            default_view_version=data_model.version or "0.1.0",
        )
        return cls(schema, metadata)

    @classmethod
    def from_directory(cls, directory: str | Path) -> "DMSImporter":
        return cls(DMSSchema.from_directory(directory))

    @classmethod
    def from_zip_file(cls, zip_file: str | Path) -> "DMSImporter":
        if Path(zip_file).suffix != ".zip":
            raise ValueError("File extension is not .zip")
        return cls(DMSSchema.from_zip(zip_file))

    @overload
    def to_rules(self, errors: Literal["raise"], role: RoleTypes | None = None, is_reference: bool = False) -> Rules:
        ...

    @overload
    def to_rules(
        self, errors: Literal["continue"] = "continue", role: RoleTypes | None = None, is_reference: bool = False
    ) -> tuple[Rules | None, IssueList]:
        ...

    def to_rules(
        self,
        errors: Literal["raise", "continue"] = "continue",
        role: RoleTypes | None = None,
        is_reference: bool = False,
    ) -> tuple[Rules | None, IssueList] | Rules:
        if role is RoleTypes.domain_expert:
            raise ValueError(f"Role {role} is not supported for DMSImporter")
        issue_list = IssueList()
        data_model = self.schema.data_models[0]

        container_by_id = {container.as_id(): container for container in self.schema.containers}

        properties = SheetList[DMSProperty]()
        for view in self.schema.views:
            class_entity = ClassEntity(prefix=view.space, suffix=view.external_id, version=view.version)
            for prop_id, prop in (view.properties or {}).items():
                if isinstance(prop, dm.MappedPropertyApply):
                    if prop.container not in container_by_id:
                        raise ValueError(f"Container {prop.container} not found")
                    container = container_by_id[prop.container]
                    if prop.container_property_identifier not in container.properties:
                        raise ValueError(
                            f"Property {prop.container_property_identifier} not found "
                            f"in container {container.external_id}"
                        )
                    container_prop = container.properties[prop.container_property_identifier]

                    index: list[str] = []
                    for index_name, index_obj in (container.indexes or {}).items():
                        if isinstance(index_obj, BTreeIndex | InvertedIndex) and prop_id in index_obj.properties:
                            index.append(index_name)
                    unique_constraints: list[str] = []
                    for constraint_name, constraint_obj in (container.constraints or {}).items():
                        if isinstance(constraint_obj, dm.RequiresConstraint):
                            # This is handled in the .from_container method of DMSContainer
                            continue
                        elif (
                            isinstance(constraint_obj, dm.UniquenessConstraint) and prop_id in constraint_obj.properties
                        ):
                            unique_constraints.append(constraint_name)
                        elif isinstance(constraint_obj, dm.UniquenessConstraint):
                            # This does not apply to this property
                            continue
                        else:
                            raise NotImplementedError(f"Constraint type {type(constraint_obj)} not implemented")

                    if isinstance(container_prop.type, dm.DirectRelation):
                        direct_value_type: str | ViewEntity | DMSValueType
                        if prop.source is None:
                            issue_list.append(
                                issues.importing.UnknownValueTypeWarning(class_entity.versioned_id, prop_id)
                            )
                            direct_value_type = ViewPropEntity(prefix=Undefined, suffix=Unknown)
                        else:
                            direct_value_type = ViewPropEntity.from_id(prop.source)

                        dms_property = DMSProperty(
                            class_=class_entity,
                            property_=prop_id,
                            description=prop.description,
                            value_type=cast(ViewPropEntity | DMSValueType, direct_value_type),
                            relation="direct",
                            nullable=container_prop.nullable,
                            default=container_prop.default_value,
                            is_list=False,
                            container=ContainerEntity.from_id(container.as_id()),
                            container_property=prop.container_property_identifier,
                            view=ViewEntity.from_id(view.as_id()),
                            view_property=prop_id,
                            index=index or None,
                            constraint=unique_constraints or None,
                        )
                    else:
                        dms_property = DMSProperty(
                            class_=ClassEntity(prefix=view.space, suffix=view.external_id, version=view.version),
                            property_=prop_id,
                            description=prop.description,
                            value_type=cast(ViewPropEntity | DMSValueType, container_prop.type._type),
                            nullable=container_prop.nullable,
                            is_list=container_prop.type.is_list,
                            default=container_prop.default_value,
                            container=ContainerEntity.from_id(container.as_id()),
                            container_property=prop.container_property_identifier,
                            view=ViewEntity.from_id(view.as_id()),
                            view_property=prop_id,
                            index=index or None,
                            constraint=unique_constraints or None,
                        )
                elif isinstance(prop, dm.MultiEdgeConnectionApply):
                    view_entity = ViewPropEntity.from_id(prop.source)
                    dms_property = DMSProperty(
                        class_=ClassEntity(prefix=view.space, suffix=view.external_id, version=view.version),
                        property_=prop_id,
                        relation="multiedge",
                        description=prop.description,
                        value_type=view_entity,
                        view=ViewEntity.from_id(view.as_id()),
                        view_property=prop_id,
                    )
                else:
                    raise NotImplementedError(f"Property type {type(prop)} not implemented")

                properties.append(dms_property)

        data_model_view_ids: set[dm.ViewId] = {
            view.as_id() if isinstance(view, dm.View | dm.ViewApply) else view for view in data_model.views or []
        }

        dms_rules = DMSRules(
            metadata=self.metadata or DMSMetadata.from_data_model(data_model),
            properties=properties,
            containers=SheetList[DMSContainer](
                data=[DMSContainer.from_container(container) for container in self.schema.containers]
            ),
            views=SheetList[DMSView](data=[DMSView.from_view(view, data_model_view_ids) for view in self.schema.views]),
            is_reference=is_reference,
        )
        output_rules: Rules
        if role is RoleTypes.information_architect:
            output_rules = dms_rules.as_information_architect_rules()
        else:
            output_rules = dms_rules
        if errors == "raise":
            return output_rules
        else:
            return output_rules, issue_list



import re
import warnings
from typing import Literal, overload

from cognite.neat.exceptions import wrangle_warnings
from cognite.neat.rules import exceptions
from cognite.neat.rules.models._rules import InformationRules
from cognite.neat.rules.models._rules._types._base import DMS_PROPERTY_ID_COMPLIANCE_REGEX, VIEW_ID_COMPLIANCE_REGEX


@overload
def are_entity_names_dms_compliant(rules: InformationRules, return_report: Literal[True]) -> tuple[bool, list[dict]]:
    ...


@overload
def are_entity_names_dms_compliant(rules: InformationRules, return_report: Literal[False] = False) -> bool:
    ...


def are_entity_names_dms_compliant(
    rules: InformationRules, return_report: bool = False
) -> bool | tuple[bool, list[dict]]:
    """Check if data model definitions are valid."""

    flag: bool = True
    with warnings.catch_warnings(record=True) as validation_warnings:
        for class_ in rules.classes:
            if not re.match(VIEW_ID_COMPLIANCE_REGEX, class_.class_.suffix):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Class", class_.class_.versioned_id, f"[Classes/Class/{class_.class_.versioned_id}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

        for row, property_ in enumerate(rules.properties):
            # check class id which would resolve as view/container id
            if not re.match(VIEW_ID_COMPLIANCE_REGEX, property_.class_.suffix):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Class", property_.class_.versioned_id, f"[Properties/Class/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

            # check property id which would resolve as view/container id
            if not re.match(DMS_PROPERTY_ID_COMPLIANCE_REGEX, property_.property_):
                warnings.warn(
                    exceptions.EntityIDNotDMSCompliant(
                        "Property", property_.property_, f"[Properties/Property/{row}]"
                    ).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )
                flag = False

    if return_report:
        return flag, wrangle_warnings(validation_warnings)
    else:
        return flag


@overload
def are_properties_redefined(rules: InformationRules, return_report: Literal[True]) -> tuple[bool, list[dict]]:
    ...


@overload
def are_properties_redefined(rules: InformationRules, return_report: Literal[False] = False) -> bool:
    ...


def are_properties_redefined(rules: InformationRules, return_report: bool = False) -> bool | tuple[bool, list[dict]]:
    flag: bool = False
    with warnings.catch_warnings(record=True) as validation_warnings:
        analyzed_properties: dict[str, list[str]] = {}
        for property_ in rules.properties:
            if property_.property_ not in analyzed_properties:
                analyzed_properties[property_.property_] = [property_.class_.versioned_id]
            elif property_.class_ in analyzed_properties[property_.property_]:
                flag = True
                warnings.warn(
                    exceptions.PropertyRedefined(property_.property_, property_.class_.versioned_id).message,
                    category=exceptions.EntityIDNotDMSCompliant,
                    stacklevel=2,
                )

            else:
                analyzed_properties[property_.property_].append(property_.class_.versioned_id)

    if return_report:
        return flag, wrangle_warnings(validation_warnings)
    else:
        return flag


def property_ids_camel_case_compliant(rules) -> bool | tuple[bool, list[dict]]:
    raise NotImplementedError()


def class_id_pascal_case_compliant(rules) -> bool | tuple[bool, list[dict]]:
    raise NotImplementedError()



import typing
from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar, TypeVar

from pydantic import BaseModel, ConfigDict

from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigs


class Config(BaseModel):
    ...


class Configurable(BaseModel):
    name: str
    value: str | None = None
    label: str | None = None
    type: str | None = None  # string , secret , number , boolean , json , multi_select , single_select
    required: bool = False
    options: list[str] | None = None


class DataContract(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


T_Input = TypeVar("T_Input", bound=DataContract)
T_Input1 = TypeVar("T_Input1", bound=DataContract)
T_Input2 = TypeVar("T_Input2", bound=DataContract)
T_Output = TypeVar("T_Output", bound=DataContract)


class Step(ABC):
    description: str = ""
    category: str = "default"
    configurables: ClassVar[list[Configurable]] = []
    scope: str = "core_global"
    metrics: NeatMetricsCollector | None = None
    workflow_configs: WorkflowConfigs | None = None
    version: str = "1.0.0"  # version of the step. All alpha versions considered as experimental
    source: str = (
        "cognite"  # source of the step , can be source identifier or url , for instance github url for instance.
    )
    docs_url: str = "https://cognite-neat.readthedocs-hosted.com/en/latest/"  # url to the documentation of the step

    def __init__(self, data_store_path: Path | None = None):
        self.log: bool = False
        self.configs: dict[str, str] = {}
        self.complex_configs: dict[
            str, typing.Any
        ] = {}  # complex configs are meant for more complex configurations. Value can be any type.
        self.workflow_id: str = ""
        self.workflow_run_id: str = ""
        self.data_store_path = Path(data_store_path) if data_store_path is not None else Path.cwd()

    @property
    def _not_configured_message(self) -> str:
        return f"Step {type(self).__name__} has not been configured."

    def set_metrics(self, metrics: NeatMetricsCollector | None):
        self.metrics = metrics

    def set_workflow_configs(self, configs: WorkflowConfigs | None):
        self.workflow_configs = configs

    def set_workflow_metadata(self, workflow_id: str, workflow_run_id: str):
        self.workflow_id = workflow_id
        self.workflow_run_id = workflow_run_id

    def configure(self, configs: dict[str, str], complex_configs: dict[str, typing.Any] | None = None):
        if complex_configs is None:
            complex_configs = {}
        self.configs = configs
        self.complex_configs = complex_configs

    def set_flow_context(self, context: dict[str, DataContract]):
        self.flow_context = context

    @abstractmethod
    def run(self, *input_data: DataContract) -> DataContract | tuple[FlowMessage, DataContract] | FlowMessage:
        ...



from abc import ABC
from dataclasses import dataclass
from pathlib import Path

from .base import NeatValidationError, ValidationWarning

__all__ = [
    "FileReadWarning",
    "InvalidFileFormatWarning",
    "UnsupportedSpecWarning",
    "UnknownItemWarning",
    "BugInImporterWarning",
    "FileReadError",
    "FileNotFoundError",
    "FileNotAFileError",
]


@dataclass(frozen=True)
class FileReadError(NeatValidationError, ABC):
    description = "An error was raised during reading."
    fix = "No fix is available."

    filepath: Path

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["filepath"] = str(self.filepath)
        return output


@dataclass(frozen=True)
class FileReadWarning(ValidationWarning, ABC):
    description = "A warning was raised during reading."
    fix = "No fix is available."

    filepath: Path

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["filepath"] = str(self.filepath)
        return output


@dataclass(frozen=True)
class InvalidFileFormatWarning(FileReadWarning):
    description = "The file format is invalid"
    fix = "Check if the file format is supported."

    reason: str

    def message(self) -> str:
        return f"Skipping invalid file {self.filepath.name}: {self.reason}"

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class UnsupportedSpecWarning(FileReadWarning):
    description = "The spec in the file is not supported"
    fix = "Change to an supported spec"

    spec_name: str
    version: str | None = None

    def message(self) -> str:
        if self.version:
            suffix = f"{self.spec_name} v{self.version} is not supported."
        else:
            suffix = f"{self.spec_name} is not supported."
        return f"Skipping file {self.filepath.name}: {suffix}. {self.fix}"

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["spec_name"] = self.spec_name
        output["version"] = self.version
        return output


@dataclass(frozen=True)
class UnknownItemWarning(FileReadWarning):
    description = "The file is missing an implementation"
    fix = "Check if the file is supported. If not, contact the neat team on Github."

    reason: str

    def message(self) -> str:
        return f"Skipping file {self.filepath.name}: {self.reason}. {self.fix}"

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["reason"] = self.reason
        return output


@dataclass(frozen=True)
class BugInImporterWarning(FileReadWarning):
    description = "A bug was raised during reading."
    fix = "No fix is available. Contact the neat team on Github"

    importer_name: str
    error: str

    def message(self) -> str:
        return f"Bug in importer {self.importer_name} when reading {self.filepath.name}: {self.error}"

    def dump(self) -> dict[str, str | None]:
        output = super().dump()
        output["importer_name"] = self.importer_name
        output["error"] = self.error
        return output


@dataclass(frozen=True)
class FileNotFoundError(FileReadError):
    description = "The file was not found"
    fix = "Check if the file exists."

    def message(self) -> str:
        return f"File {self.filepath} was not found. {self.fix}"


@dataclass(frozen=True)
class FileNotAFileError(FileReadError):
    description = "The file is not a file"
    fix = "Check if the file exists."

    def message(self) -> str:
        return f"{self.filepath} is not a file. {self.fix}"


@dataclass(frozen=True)
class InvalidFileFormatError(FileReadError):
    description = "The file is not in the expected format"
    fix = "Check if the file is in the expected format"

    expected_format: list[str]

    def message(self) -> str:
        return f"{self.filepath} is not in the expected format. Expected format: {self.expected_format}."


@dataclass(frozen=True)
class NoFilesFoundError(FileReadError):
    description = "No files were found in the directory"
    fix = "Check if the directory contains files"

    expected_formats: list[str] | None = None

    def message(self) -> str:
        if self.expected_formats:
            return f"No files were found in {self.filepath.name}. Expected format: {self.expected_formats}. {self.fix}"
        return f"No files were found in {self.filepath.name}. {self.fix}"



import re
import sys
from functools import total_ordering
from typing import Any, ClassVar, Literal, cast, overload

from cognite.client.data_classes.data_modeling import ContainerId, DataModelId, PropertyId, ViewId
from pydantic import BaseModel

from cognite.neat.rules.models.rdfpath import (
    SINGLE_PROPERTY_REGEX_COMPILED,
    AllReferences,
    RDFPath,
    SingleProperty,
    TransformationRuleType,
    parse_rule,
)

if sys.version_info >= (3, 11):
    from enum import StrEnum
    from typing import Self
else:
    from backports.strenum import StrEnum
    from typing_extensions import Self


class EntityTypes(StrEnum):
    subject = "subject"
    predicate = "predicate"
    object = "object"
    class_ = "class"
    parent_class = "parent_class"
    property_ = "property"
    object_property = "ObjectProperty"
    data_property = "DatatypeProperty"
    annotation_property = "AnnotationProperty"
    object_value_type = "object_value_type"
    data_value_type = "data_value_type"  # these are strings, floats, ...
    xsd_value_type = "xsd_value_type"
    dms_value_type = "dms_value_type"
    view = "view"
    view_prop = "view_prop"
    reference_entity = "reference_entity"
    container = "container"
    datamodel = "datamodel"
    undefined = "undefined"


# ALLOWED
ALLOWED_PATTERN = r"[^a-zA-Z0-9-_.]"

# FOR PARSING STRINGS:
PREFIX_REGEX = r"[a-zA-Z]+[a-zA-Z0-9-_.]*[a-zA-Z0-9]+"
SUFFIX_REGEX = r"[a-zA-Z0-9-_.]+[a-zA-Z0-9]|[-_.]*[a-zA-Z0-9]+"
VERSION_REGEX = r"[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?"
PROPERTY_REGEX = r"[a-zA-Z0-9][a-zA-Z0-9_-]*[a-zA-Z0-9]?"
ENTITY_ID_REGEX = rf"{PREFIX_REGEX}:({SUFFIX_REGEX})"
ENTITY_ID_REGEX_COMPILED = re.compile(rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})$")
VERSIONED_ENTITY_REGEX_COMPILED = re.compile(
    rf"^(?P<prefix>{PREFIX_REGEX}):(?P<suffix>{SUFFIX_REGEX})\(version=(?P<version>{VERSION_REGEX})\)$"
)
PROPERTY_ENTITY_REGEX_COMPILED = re.compile(
    rf"^((?P<prefix>{PREFIX_REGEX}):)?(?P<suffix>{SUFFIX_REGEX})"
    rf"\((version=(?P<version>{VERSION_REGEX}), )?property=(?P<property>{PROPERTY_REGEX})\)$"
)
CLASS_ID_REGEX = rf"(?P<{EntityTypes.class_}>{ENTITY_ID_REGEX})"
CLASS_ID_REGEX_COMPILED = re.compile(rf"^{CLASS_ID_REGEX}$")
PROPERTY_ID_REGEX = rf"\((?P<{EntityTypes.property_}>{ENTITY_ID_REGEX})\)"
VERSION_ID_REGEX = rf"\(version=(?P<version>{VERSION_REGEX})\)"
MORE_THAN_ONE_NONE_ALPHANUMERIC_REGEX = r"([_-]{2,})"
PREFIX_COMPLIANCE_REGEX = r"^([a-zA-Z]+)([a-zA-Z0-9]*[_-]{0,1}[a-zA-Z0-9_-]*)([a-zA-Z0-9]*)$"
DATA_MODEL_ID_COMPLIANCE_REGEX = r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$"
CDF_SPACE_COMPLIANCE_REGEX = (
    r"(?!^(space|cdf|dms|pg3|shared|system|node|edge)$)(^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$)"
)
VIEW_ID_COMPLIANCE_REGEX = (
    r"(?!^(Query|Mutation|Subscription|String|Int32|Int64|Int|Float32|Float64|Float|"
    r"Timestamp|JSONObject|Date|Numeric|Boolean|PageInfo|File|Sequence|TimeSeries)$)"
    r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)
DMS_PROPERTY_ID_COMPLIANCE_REGEX = (
    r"(?!^(space|externalId|createdTime|lastUpdatedTime|deletedTime|edge_id|"
    r"node_id|project_id|property_group|seq|tg_table_name|extensions)$)"
    r"(^[a-zA-Z][a-zA-Z0-9_]{0,253}[a-zA-Z0-9]?$)"
)
CLASS_ID_COMPLIANCE_REGEX = r"(?!^(Class|class)$)(^[a-zA-Z][a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"
PROPERTY_ID_COMPLIANCE_REGEX = r"^(\*)|(?!^(Property|property)$)(^[a-zA-Z][a-zA-Z0-9._-]{0,253}[a-zA-Z0-9]?$)"
VERSION_COMPLIANCE_REGEX = r"^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$"


Undefined = type(object())
Unknown = type(object())


# mypy does not like the sentinel value, and it is not possible to ignore only the line with it below.
# so we ignore all errors beyond this point.
# mypy: ignore-errors
@total_ordering
class Entity(BaseModel, arbitrary_types_allowed=True):
    """Entity is a class or property in OWL/RDF sense."""

    type_: ClassVar[EntityTypes] = EntityTypes.undefined
    prefix: str | Undefined = Undefined
    suffix: str | Unknown
    version: str | None = None
    name: str | None = None
    description: str | None = None

    def __lt__(self, other: object) -> bool:
        if type(self) is not type(other) or not isinstance(other, Entity):
            return NotImplemented
        return self.versioned_id < other.versioned_id

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other) or not isinstance(other, Entity):
            return NotImplemented
        return self.versioned_id == other.versioned_id

    def __hash__(self) -> int:
        return hash(self.versioned_id)

    def as_non_versioned_entity(self) -> Self:
        return self.from_string(f"{self.prefix}:{self.suffix}")

    @property
    def id(self) -> str:
        if self.suffix is Unknown:
            return "#N/A"
        elif self.prefix is Undefined:
            return self.suffix
        else:
            return f"{self.prefix}:{self.suffix}"

    @property
    def versioned_id(self) -> str:
        if self.version is None:
            return self.id
        else:
            return f"{self.id}(version={self.version})"

    @property
    def space(self) -> str:
        """Returns entity space in CDF."""
        return self.prefix

    @property
    def external_id(self) -> str:
        """Returns entity external id in CDF."""
        return self.suffix

    def __repr__(self):
        return self.versioned_id

    def __str__(self):
        return self.versioned_id

    @classmethod
    def from_string(cls, entity_string: str, base_prefix: str | None = None) -> Self:
        if entity_string == "#N/A":
            return cls(prefix=Undefined, suffix=Unknown)
        elif result := VERSIONED_ENTITY_REGEX_COMPILED.match(entity_string):
            return cls(
                prefix=result.group("prefix"),
                suffix=result.group("suffix"),
                version=result.group("version"),
            )
        elif result := ENTITY_ID_REGEX_COMPILED.match(entity_string):
            return cls(prefix=result.group("prefix"), suffix=result.group("suffix"))
        elif base_prefix and re.match(SUFFIX_REGEX, entity_string) and re.match(PREFIX_REGEX, base_prefix):
            return cls(prefix=base_prefix, suffix=entity_string)
        else:
            raise ValueError(f"{cls.__name__} is expected to be prefix:suffix, got {entity_string}")

    @classmethod
    def from_list(cls, entity_strings: list[str], base_prefix: str | None = None) -> list[Self]:
        return [
            cls.from_string(entity_string=entity_string, base_prefix=base_prefix) for entity_string in entity_strings
        ]


class ContainerEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.container

    @classmethod
    def from_raw(cls, value: Any) -> "ContainerEntity":
        if not value:
            return ContainerEntity(prefix=Undefined, suffix=value)
        elif isinstance(value, ContainerEntity):
            return value

        if ENTITY_ID_REGEX_COMPILED.match(value):
            return ContainerEntity.from_string(entity_string=value)
        else:
            return ContainerEntity(prefix=Undefined, suffix=value)

    @classmethod
    def from_id(cls, container_id: ContainerId) -> "ContainerEntity":
        return ContainerEntity(prefix=container_id.space, suffix=container_id.external_id)

    def as_id(self, default_space: str | None) -> ContainerId:
        if self.space is Undefined and default_space is None:
            raise ValueError("Space is Undefined! Set default_space!")

        if self.space is Undefined:
            return ContainerId(space=cast(str, default_space), external_id=self.external_id)
        else:
            return ContainerId(space=self.space, external_id=self.external_id)


class ViewEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.view

    @classmethod
    def from_raw(cls, value: Any) -> "ViewEntity":
        if not value:
            return ViewEntity(prefix=Undefined, suffix=value)
        elif isinstance(value, ViewEntity):
            return value

        if ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
            return cls.from_string(entity_string=value)
        else:
            return cls(prefix=Undefined, suffix=value)

    @classmethod
    def from_id(cls, view_id: ViewId) -> Self:
        return cls(prefix=view_id.space, suffix=view_id.external_id, version=view_id.version)

    @overload
    def as_id(
        self,
        allow_none: Literal[False] = False,
        default_space: str | None = None,
        default_version: str | None = None,
    ) -> ViewId:
        ...

    @overload
    def as_id(
        self,
        allow_none: Literal[True],
        default_space: str | None = None,
        default_version: str | None = None,
    ) -> ViewId | None:
        ...

    def as_id(
        self,
        allow_none: bool = False,
        default_space: str | None = None,
        default_version: str | None = None,
    ) -> ViewId | None:
        if self.suffix is Unknown and allow_none:
            return None
        elif self.suffix is Unknown:
            raise ValueError("suffix is Unknown and cannot be converted to ViewId")
        space = default_space if self.space is Undefined else self.space
        version = self.version or default_version

        if space is None or space is Undefined:
            raise ValueError("space is required")
        if version is None:
            raise ValueError("version is required")
        return ViewId(space=space, external_id=self.external_id, version=version)


class ViewPropEntity(ViewEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.view_prop
    property_: str | None = None

    @classmethod
    def from_raw(cls, value: Any) -> "ViewPropEntity":
        if not value:
            return ViewPropEntity(prefix=Undefined, suffix=value)
        elif isinstance(value, ViewPropEntity):
            return value
        elif isinstance(value, ViewEntity):
            return cls(prefix=value.prefix, suffix=value.suffix, version=value.version)

        if result := PROPERTY_ENTITY_REGEX_COMPILED.match(value):
            return cls(
                prefix=result.group("prefix") or Undefined,
                suffix=result.group("suffix"),
                version=result.group("version"),
                property_=result.group("property"),
            )
        elif ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
            return cls.from_string(entity_string=value)
        else:
            return cls(prefix=Undefined, suffix=value)

    @classmethod
    def from_prop_id(cls, prop_id: PropertyId) -> "ViewPropEntity":
        return ViewPropEntity(
            prefix=prop_id.source.space,
            suffix=prop_id.source.external_id,
            version=prop_id.source.version,
            property_=prop_id.property,
        )

    def as_prop_id(self, default_space: str | None = None, default_version: str | None = None) -> PropertyId:
        if self.property_ is None:
            raise ValueError("property is required to create PropertyId")
        return PropertyId(source=self.as_id(False, default_space, default_version), property=self.property_)

    @property
    def versioned_id(self) -> str:
        if self.version is None and self.property_ is None:
            output = self.id
        elif self.version is None:
            output = f"{self.id}(property={self.property_})"
        elif self.property_ is None:
            output = f"{self.id}(version={self.version})"
        else:
            output = f"{self.id}(version={self.version}, property={self.property_})"
        return output


class DataModelEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.datamodel

    @classmethod
    def from_raw(cls, value: Any) -> "DataModelEntity":
        if not value:
            return DataModelEntity(prefix=Undefined, suffix=value)
        elif isinstance(value, DataModelEntity):
            return value

        if ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
            return DataModelEntity.from_string(entity_string=value)
        else:
            return DataModelEntity(prefix=Undefined, suffix=value)

    @classmethod
    def from_id(cls, data_model_id: DataModelId) -> "DataModelEntity":
        return DataModelEntity(
            prefix=data_model_id.space, suffix=data_model_id.external_id, version=data_model_id.version
        )

    def as_id(self, default_space: str | None = None, default_version: str | None = None) -> DataModelId:
        space = default_space if self.space is Undefined else self.space
        version = self.version or default_version

        if space is None or space is Undefined:
            raise ValueError("space is required")

        return DataModelId(space=space, external_id=self.external_id, version=version)


class ClassEntity(Entity):
    type_: ClassVar[EntityTypes] = EntityTypes.class_

    @classmethod
    def from_raw(cls, value: Any) -> Self:
        if not value:
            return cls(prefix=Undefined, suffix=value)
        elif isinstance(value, cls):
            return value

        if ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
            return cls.from_string(entity_string=value)
        else:
            return cls(prefix=Undefined, suffix=value)

    @classmethod
    def from_view_id(cls, view_id: ViewId) -> Self:
        return cls(prefix=view_id.space, suffix=view_id.external_id, version=view_id.version)

    @property
    def view_id(self) -> ViewId:
        return ViewId(space=self.space, external_id=self.external_id, version=self.version)


class ParentClassEntity(ClassEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.parent_class

    @classmethod
    def from_raw(cls, value: Any) -> Self:
        if isinstance(value, ClassEntity):
            return cls(prefix=value.prefix, suffix=value.suffix, version=value.version)
        return super().from_raw(value)

    def as_class_entity(self) -> ClassEntity:
        return ClassEntity(prefix=self.prefix, suffix=self.suffix, version=self.version)


class ReferenceEntity(ViewPropEntity):
    type_: ClassVar[EntityTypes] = EntityTypes.reference_entity

    @classmethod
    def from_raw(cls, value: Any) -> "ReferenceEntity":
        if not value:
            return ReferenceEntity(prefix=Undefined, suffix=value)
        elif isinstance(value, ReferenceEntity):
            return value
        elif isinstance(value, ViewEntity):
            return cls(prefix=value.prefix, suffix=value.suffix, version=value.version)

        if result := PROPERTY_ENTITY_REGEX_COMPILED.match(value):
            return cls(
                prefix=result.group("prefix") or Undefined,
                suffix=result.group("suffix"),
                version=result.group("version"),
                property_=result.group("property"),
            )
        elif ENTITY_ID_REGEX_COMPILED.match(value) or VERSIONED_ENTITY_REGEX_COMPILED.match(value):
            return cls.from_string(entity_string=value)
        elif result := SINGLE_PROPERTY_REGEX_COMPILED.match(value):
            return cls.from_rdfpath(value)
        else:
            return cls(prefix=Undefined, suffix=value)

    @classmethod
    def from_rdfpath(cls, rdfpath: RDFPath | str) -> "ReferenceEntity":
        if isinstance(rdfpath, str):
            rdfpath = parse_rule(rdfpath, TransformationRuleType.rdfpath)

        if isinstance(rdfpath.traversal, AllReferences):
            return cls(prefix=rdfpath.traversal.class_.prefix, suffix=rdfpath.traversal.class_.suffix)
        elif isinstance(rdfpath.traversal, SingleProperty):
            return cls(
                prefix=rdfpath.traversal.class_.prefix,
                suffix=rdfpath.traversal.class_.suffix,
                property_=rdfpath.traversal.property.suffix,
            )
        else:
            raise ValueError(f"Invalid RDFPath traversal type: {rdfpath.traversal}")

    def as_rdfpath(self, default_prefix: str | None = None) -> RDFPath:
        prefix = self.prefix if self.prefix is not Undefined else default_prefix

        if prefix is None:
            raise ValueError("prefix is required")

        if self.property_:
            return parse_rule(f"{prefix}:{self.suffix}({self.prefix}:{self.property_})", TransformationRuleType.rdfpath)
        else:
            return parse_rule(f"{self.prefix}:{self.suffix}", TransformationRuleType.rdfpath)

    def as_class_entity(self) -> ClassEntity:
        return ClassEntity(prefix=self.prefix, suffix=self.suffix, version=self.version)



import json
import warnings
from pathlib import Path
from typing import Literal, get_args

import yaml

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models._rules.base import RoleTypes

from ._base import BaseExporter


class YAMLExporter(BaseExporter[str]):
    """Export rules to YAML.

    Args:
        files: The number of files to output. Defaults to "single".
        output: The format to output the rules. Defaults to "yaml".
        output_role: The role to output the rules for. Defaults to None, which means that the role
            of the input rules will be used.

    The following formats are available:

    - "single": A single YAML file will contain the entire rules.

    .. note::

        YAML files are typically used for storing rules when checked into version control systems, e.g., git-history.
        The advantage of using YAML files over Excel is that tools like git can show the differences between different
        versions of the rules.

    """

    Files = Literal["single"]
    Format = Literal["yaml", "json"]

    file_option = get_args(Files)
    format_option = get_args(Format)

    def __init__(self, files: Files = "single", output: Format = "yaml", output_role: RoleTypes | None = None):
        if files not in self.file_option:
            raise ValueError(f"Invalid files: {files}. Valid options are {self.file_option}")
        if output not in self.format_option:
            raise ValueError(f"Invalid output: {output}. Valid options are {self.format_option}")
        self.files = files
        self.output = output
        self.output_role = output_role

    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        """Exports transformation rules to YAML/JSON file(s)."""
        if self.files == "single":
            if filepath.suffix != f".{self.output}":
                warnings.warn(f"File extension is not .{self.output}, adding it to the file name", stacklevel=2)
                filepath = filepath.with_suffix(f".{self.output}")
            filepath.write_text(self.export(rules), encoding=self._encoding, newline=self._new_line)
        else:
            raise NotImplementedError(f"Exporting to {self.files} files is not supported")

    def export(self, rules: Rules) -> str:
        """Export rules to YAML (or JSON) format.

        Args:
            rules: The rules to be exported.

        Returns:
            str: The rules in YAML (or JSON) format.
        """
        rules = self._convert_to_output_role(rules, self.output_role)
        # model_dump_json ensures that the output is in JSON format,
        # if we don't do this, we will get Enums and other types that are not serializable to YAML
        json_output = rules.model_dump_json()
        if self.output == "json":
            return json_output
        elif self.output == "yaml":
            return yaml.safe_dump(json.loads(json_output))
        else:
            raise ValueError(f"Invalid output: {self.output}. Valid options are {self.format_option}")



"""
This is a pydantic validation implementation of the DTDL v2 and v3 specifications.

The specs are taken from:

 * Spec v2:  https://github.com/Azure/opendigitaltwins-dtdl/blob/master/DTDL/v2/DTDL.v2.md
 * Spec v3: https://github.com/Azure/opendigitaltwins-dtdl/blob/master/DTDL/v3/DTDL.v3.md
"""

import re
import warnings
from abc import ABC
from typing import TYPE_CHECKING, Any, ClassVar, Literal, TypeAlias

from pydantic import BaseModel, Field, field_validator, model_serializer, model_validator

from cognite.neat.rules.models._rules._types import ClassEntity

if TYPE_CHECKING:
    from pydantic.type_adapter import IncEx


# Regex is from the spec: https://github.com/Azure/opendigitaltwins-dtdl/blob/master/DTMI/README.md#validation-regular-expressions
_DTMI_REGEX = (
    r"^dtmi:(?:_+[A-Za-z0-9]|[A-Za-z])(?:[A-Za-z0-9_]*[A-Za-z0-9])?(?::(?:_+[A-Za-z0-9]|[A-Za-z])"
    r"(?:[A-Za-z0-9_]*[A-Za-z0-9])?)*(?:;[1-9][0-9]{0,8}(?:\.[1-9][0-9]{0,5})?)?$"
)

_DTMI_COMPILED = re.compile(_DTMI_REGEX)


class DTMI(BaseModel):
    scheme: ClassVar[str] = "dtmi"
    path: list[str]
    version: str

    def as_class_id(self) -> ClassEntity:
        return ClassEntity(prefix="_".join(self.path[:-1]), suffix=self.path[-1], version=self.version)

    def __hash__(self) -> int:
        return hash(self.to_string())

    def __repr__(self) -> str:
        return self.to_string()

    @model_validator(mode="before")
    def from_string(cls, value: Any) -> Any:
        if not isinstance(value, str):
            return value
        if not _DTMI_COMPILED.match(value):
            raise ValueError(f"Invalid DTMI {value}")
        value = value.removeprefix(cls.scheme + ":")
        path_str, version = value.split(";", 1)
        return dict(path=path_str.split(":"), version=version)

    @model_serializer
    def to_string(self) -> str:
        return f"{self.scheme}:{':'.join(self.path)};{self.version}"

    if TYPE_CHECKING:
        # Ensure type checkers works correctly, ref
        # https://docs.pydantic.dev/latest/concepts/serialization/#overriding-the-return-type-when-dumping-a-model
        def model_dump(  # type: ignore[override]
            self,
            *,
            mode: Literal["json", "python"] | str = "python",
            include: IncEx | None = None,
            exclude: IncEx | None = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
        ) -> str:
            ...


IRI: TypeAlias = str


class Unit(BaseModel, ABC):
    value: str
    semantic_type: str | None = Field(None, alias="semanticType")
    unit_type: str | None = Field(None, alias="unitType")

    def __hash__(self) -> int:
        return hash(self.to_string())

    def __repr__(self) -> str:
        return self.to_string()

    @model_validator(mode="before")
    def from_string(cls, value: Any) -> Any:
        if not isinstance(value, str):
            return value
        from ._unit_lookup import ENTRY_BY_UNIT

        if value not in ENTRY_BY_UNIT:
            return dict(unit=value)
        entry = ENTRY_BY_UNIT[value]
        return dict(
            value=value,
            semanticType=entry.semantic_type,
            unitType=entry.unit_type,
        )

    @model_serializer
    def to_string(self) -> str:
        return self.value

    if TYPE_CHECKING:
        # Ensure type checkers works correctly, ref
        # https://docs.pydantic.dev/latest/concepts/serialization/#overriding-the-return-type-when-dumping-a-model
        def model_dump(  # type: ignore[override]
            self,
            *,
            mode: Literal["json", "python"] | str = "python",
            include: IncEx | None = None,
            exclude: IncEx | None = None,
            by_alias: bool = False,
            exclude_unset: bool = False,
            exclude_defaults: bool = False,
            exclude_none: bool = False,
            round_trip: bool = False,
            warnings: bool = True,
        ) -> str:
            ...


class DTDLBase(BaseModel, ABC):
    type: ClassVar[str]
    spec_version: ClassVar[frozenset[str]]
    id_: DTMI | None = Field(None, alias="@id")
    comment: str | None = None
    display_name: str | None = Field(None, alias="displayName")
    description: str | None = None


PrimitiveSchema: TypeAlias = Literal[
    "boolean", "date", "dateTime", "double", "duration", "float", "integer", "long", "string", "time"
]


class DTDLBaseWithName(DTDLBase, ABC):
    name: str


class DTDLBaseWithSchema(DTDLBaseWithName, ABC):
    schema_: "Schema | DTMI | None" = Field(None, alias="schema")  # type: ignore[assignment]

    @field_validator("schema_", mode="before")
    def select_schema_type(cls, value: Any) -> Any:
        if isinstance(value, dict) and (type_ := value.get("@type")):
            context = Interface.default_context
            spec_version = context.rsplit(";", maxsplit=1)[1]
            try:
                cls_by_type = DTDL_CLS_BY_TYPE_BY_SPEC[spec_version]
            except KeyError:
                raise ValueError(f"DTDL Spec v{spec_version} is not supported: {context}") from None
            if isinstance(type_, str) and (cls_ := cls_by_type.get(type_)) is not None:
                return cls_.model_validate(value)
            elif isinstance(type_, list) and len(type_) == 2 and (cls_ := cls_by_type.get(type_[0])) is not None:
                # In the spec v2, the type of Telemetry and Property can be a list of two strings,
                # [[Telemetry|Property, "Semantic Type"].
                from ._unit_lookup import UNIT_TYPE_BY_SEMANTIC_TYPE

                if unit_type := UNIT_TYPE_BY_SEMANTIC_TYPE.get(type_[1]) and "unit" in value:
                    value["unit"] = {
                        "value": value["unit"],
                        "semanticType": type_[1],
                        "unitType": unit_type,
                    }
                return cls_.model_validate(value)
        return value


class DTDLField(DTDLBaseWithSchema):
    type = "Field"
    spec_version = frozenset(["2", "3"])


class Object(DTDLBase):
    type = "Object"
    spec_version = frozenset(["2", "3"])
    fields: list[DTDLField] | None = None


class MapKey(DTDLBaseWithName):
    type = "MapKey"
    spec_version = frozenset(["2", "3"])
    schema_: str = Field(alias="schema")


class MapValue(DTDLBaseWithSchema):
    type = "MapValue"
    spec_version = frozenset(["2", "3"])


class Map(DTDLBase):
    type = "Map"
    spec_version = frozenset(["2", "3"])
    map_key: MapKey = Field(alias="mapKey")
    map_value: MapValue = Field(alias="mapValue")


class EnumValue(DTDLBaseWithName):
    type = "EnumValue"
    spec_version = frozenset(["2", "3"])
    enum_value: str = Field(alias="enumValue")


class Enum(DTDLBase):
    type = "Enum"
    spec_version = frozenset(["2", "3"])
    enum_values: list[EnumValue] = Field(alias="enumValues")
    value_schema: PrimitiveSchema = Field(alias="valueSchema")


class Array(DTDLBase):
    type = "Array"
    spec_version = frozenset(["2", "3"])
    element_schema: "Schema" = Field(alias="elementSchema")


ComplexSchema: TypeAlias = Array | Enum | Map | Object

Schema: TypeAlias = PrimitiveSchema | ComplexSchema


class Component(DTDLBaseWithSchema):
    type = "Component"
    spec_version = frozenset(["2", "3"])
    schema_: "Interface | DTMI" = Field(alias="schema")  # type: ignore[assignment]


class Property(DTDLBaseWithSchema):
    type = "Property"
    spec_version = frozenset(["3"])
    writable: bool | None = None


class PropertyV2(Property):
    spec_version = frozenset(["2"])
    unit: Unit | None = None


class Relationship(DTDLBaseWithName):
    type = "Relationship"
    spec_version = frozenset(["2", "3"])
    min_multiplicity: int | None = Field(None, alias="minMultiplicity", le=0, ge=0)
    max_multiplicity: int | None = Field(None, alias="maxMultiplicity", ge=1)
    properties: list[Property] | None = None
    target: DTMI | None = None
    writable: bool | None = None


class CommandRequest(DTDLBaseWithSchema):
    type = "CommandRequest"
    spec_version = frozenset(["3"])


class CommandResponse(DTDLBaseWithSchema):
    type = "CommandResponse"
    spec_version = frozenset(["3"])


class CommandPayload(DTDLBaseWithSchema):
    type = "CommandPayload"
    spec_version = frozenset(["2"])


class Command(DTDLBaseWithSchema):
    type = "Command"
    spec_version = frozenset(["3"])
    request: CommandRequest | None = None
    response: CommandResponse | None = None


class CommandV2(DTDLBaseWithSchema):
    type = "Command"
    spec_version = frozenset(["2"])
    request: CommandPayload | None = None
    response: CommandPayload | None = None


class Telemetry(DTDLBaseWithSchema):
    type = "Telemetry"
    spec_version = frozenset(["3"])


class TelemetryV2(Telemetry):
    spec_version = frozenset(["2"])
    unit: Unit | None = None


class Interface(DTDLBase):
    type = "Interface"
    spec_version = frozenset(["2", "3"])
    default_context: ClassVar[IRI] = Field(
        "dtmi:dtdl:context;3",
        description="This can be set directly on the class to change the "
        "default context used when parsing a document.",
    )
    id_: DTMI = Field(alias="@id")  # type: ignore[assignment]
    context: IRI | None = Field(alias="@context")
    extends: list[DTMI] | None = None
    contents: list[Command | Component | Property | Relationship | Telemetry | DTMI | CommandV2] | None = None
    schemas: list[Array | Enum | Map | Object] | None = None

    @field_validator("context", mode="before")
    def list_to_string(cls, value: Any) -> Any:
        if isinstance(value, list) and len(value) == 1:
            return value[0]
        return value

    @field_validator("contents", "schemas", mode="before")
    def select_content_type(cls, value: Any, info) -> Any:
        if not isinstance(value, list):
            return value
        context = info.data.get("@context", cls.default_context)
        spec_version = context.rsplit(";", maxsplit=1)[1]
        try:
            cls_by_type = DTDL_CLS_BY_TYPE_BY_SPEC[spec_version]
        except KeyError:
            raise ValueError(f"DTDL Spec v{spec_version} is not supported: {context}") from None
        output: list[DTDLBase] = []
        for item in value:
            if isinstance(item, dict) and (type_ := item.get("@type")):
                if isinstance(type_, str) and (cls_ := cls_by_type.get(type_)) is not None:
                    item = cls_.model_validate(item)
                elif isinstance(type_, list) and len(type_) == 2 and (cls_ := cls_by_type.get(type_[0])) is not None:
                    # In the spec v2, the type of Telemetry and Property can be a list of two strings,
                    # [[Telemetry|Property, "Semantic Type"].
                    from ._unit_lookup import UNIT_TYPE_BY_SEMANTIC_TYPE

                    if (unit_type := UNIT_TYPE_BY_SEMANTIC_TYPE.get(type_[1])) and (unit := item.get("unit")):
                        item["unit"] = {
                            "value": unit,
                            "semanticType": type_[1],
                            "unitType": unit_type,
                        }
                    item = cls_.model_validate(item)
            else:
                warnings.warn(f"Invalid item {item} in {cls.__name__}.contents", stacklevel=2)
            output.append(item)
        return output


DTDL_CLS_BY_TYPE_BY_SPEC: dict[str, dict[str, type[DTDLBase]]] = {}
to_check = list(DTDLBase.__subclasses__())
while to_check:
    cls = to_check.pop()
    to_check.extend(cls.__subclasses__())
    if ABC in cls.__bases__:
        continue
    for spec in cls.spec_version:
        DTDL_CLS_BY_TYPE_BY_SPEC.setdefault(spec, {})[cls.type] = cls
del cls, to_check, spec



import time
from pathlib import Path
from typing import ClassVar

from cognite.client import CogniteClient

from cognite.neat.rules import importers
from cognite.neat.rules.issues.formatters import FORMATTER_BY_NAME
from cognite.neat.rules.models._rules import RoleTypes
from cognite.neat.rules.models._rules._types import DataModelEntity, Undefined
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import MultiRuleData
from cognite.neat.workflows.steps.step_model import Configurable, Step

CATEGORY = __name__.split(".")[-1].replace("_", " ").title()

__all__ = [
    "ExcelToRules",
    "DMSToRules",
]


class ExcelToRules(Step):
    """This step import rules from the Excel file and validates it."""

    description = "This step imports rules from an excel file "
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="File name",
            value="",
            label="Full file name of the rules file in the rules folder. \
                If not provided, step will attempt to get file name from payload \
                    of 'File Uploader' step (if exist)",
        ),
        Configurable(
            name="Report formatter",
            value=next(iter(FORMATTER_BY_NAME.keys())),
            label="The format of the report for the validation of the rules",
            options=list(FORMATTER_BY_NAME),
        ),
        Configurable(
            name="Role",
            value="infer",
            label="For what role Rules are intended?",
            options=["infer", *RoleTypes.__members__.keys()],
        ),
    ]

    def run(self, flow_message: FlowMessage) -> (FlowMessage, MultiRuleData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        file_name = self.configs.get("File name", None)
        full_path = flow_message.payload.get("full_path", None) if flow_message.payload else None

        if file_name:
            rules_file_path = Path(self.data_store_path) / "rules" / file_name
        elif full_path:
            rules_file_path = full_path
        else:
            error_text = "Expected either 'File name' in the step config or 'File uploader' step uploading Excel Rules."
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        # if role is None, it will be inferred from the rules file
        role = self.configs.get("Role")
        role_enum = None
        if role != "infer" and role is not None:
            role_enum = RoleTypes[role]

        excel_importer = importers.ExcelImporter(rules_file_path)
        rules, issues = excel_importer.to_rules(role=role_enum, errors="continue")

        if rules is None:
            output_dir = self.data_store_path / Path("staging")
            report_writer = FORMATTER_BY_NAME[self.configs["Report formatter"]]()
            report_writer.write_to_file(issues, file_or_dir_path=output_dir)
            report_file = report_writer.default_file_name
            error_text = (
                "<p></p>"
                f'<a href="/data/staging/{report_file}?{time.time()}" '
                f'target="_blank">Failed to validate rules, click here for report</a>'
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        output_text = "Rules validation passed successfully!"

        return FlowMessage(output_text=output_text), MultiRuleData.from_rules(rules)


class DMSToRules(Step):
    """This step imports rules from CDF Data Model"""

    description = "This step imports rules from CDF Data Model"
    version = "private-beta"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="Data model id",
            value="",
            label="The ID of the Data Model to import. Written at 'my_space:my_data_model(version=1)'",
            type="string",
            required=True,
        ),
        Configurable(
            name="Report formatter",
            value=next(iter(FORMATTER_BY_NAME.keys())),
            label="The format of the report for the validation of the rules",
            options=list(FORMATTER_BY_NAME),
        ),
        Configurable(
            name="Role",
            value="infer",
            label="For what role Rules are intended?",
            options=["infer", *RoleTypes.__members__.keys()],
        ),
        Configurable(
            name="Reference",
            value="false",
            label="Whether the imported rules are a reference model or not. This is "
            "used when you want to extend an existing model and this is the model that is being extended from.",
            options=["true", "false"],
        ),
    ]

    def run(self, cdf_client: CogniteClient) -> (FlowMessage, MultiRuleData):  # type: ignore[syntax, override]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        datamodel_id_str = self.configs.get("Data model id")
        if datamodel_id_str is None:
            error_text = "Expected input payload to contain 'Data model id' key."
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        datamodel_entity = DataModelEntity.from_raw(datamodel_id_str)
        if datamodel_entity.space is Undefined:
            error_text = (
                f"Data model id should be in the format 'my_space:my_data_model(version=1)' "
                f"or 'my_space:my_data_model', failed to parse space from {datamodel_id_str}"
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        is_reference = self.configs.get("Reference", "false") == "true"

        dms_importer = importers.DMSImporter.from_data_model_id(cdf_client, datamodel_entity.as_id())

        # if role is None, it will be inferred from the rules file
        role = self.configs.get("Role")
        role_enum = None
        if role != "infer" and role is not None:
            role_enum = RoleTypes[role]

        rules, issues = dms_importer.to_rules(role=role_enum, errors="continue", is_reference=is_reference)

        if rules is None:
            output_dir = self.data_store_path / Path("staging")
            report_writer = FORMATTER_BY_NAME[self.configs["Report formatter"]]()
            report_writer.write_to_file(issues, file_or_dir_path=output_dir)
            report_file = report_writer.default_file_name
            error_text = (
                "<p></p>"
                f'<a href="/data/staging/{report_file}?{time.time()}" '
                f'target="_blank">Failed to validate rules, click here for report</a>'
            )
            return FlowMessage(error_text=error_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)

        output_text = "Rules import and validation passed successfully!"

        return FlowMessage(output_text=output_text), MultiRuleData.from_rules(rules)



from pathlib import Path

from ruamel.yaml import YAML

# Migration mapping

# LoadTransformationRules to ImportExcelToRules
# InstancesFromRdfFileToSourceGraph to ExtractGraphFromRdfFile
# InstancesFromRulesToSolutionGraph to ExtractGraphFromRulesInstanceSheet
# GraphCapturingSheetToGraph to ExtractGraphFromGraphCapturingSheet
# GenerateMockGraph to ExtractGraphFromMockGraph
# InstancesFromJsonToGraph to ExtractGraphFromJsonFile
# InstancesFromAvevaPiAF to ExtractGraphFromAvevaPiAssetFramework
# DexpiToGraph to ExtractGraphFromDexpiFile
# GenerateCDFAssetsFromGraph to GenerateAssetsFromGraph
# GenerateCDFRelationshipsFromGraph to GenerateRelationshipsFromGraph
# GenerateCDFNodesAndEdgesFromGraph to GenerateNodesAndEdgesFromGraph
# UploadCDFAssets to LoadAssetsToCDF
# UploadCDFRelationships to LoadRelationshipsToCDF
# UploadCDFNodes to LoadNodesToCDF
# UploadCDFEdges to LoadEdgesToCDF
# CreateCDFLabels to LoadLabelsToCDF
# OpenApiToRules to `ImportOpenApiToRules
# ArbitraryJsonYamlToRules to ImportArbitraryJsonYamlToRules
# GraphToRules to ImportGraphToRules
# OntologyToRules to ImportOntologyToRules
# GraphQLSchemaFromRules to ExportGraphQLSchemaFromRules
# OntologyFromRules to ExportOntologyFromRules
# SHACLFromRules to ExportSHACLFromRules
# GraphCaptureSpreadsheetFromRules to ExportRulesToGraphCapturingSheet
# ExcelFromRules to ExportRulesToExcel

# Define a dictionary that maps old step names to new ones
step_rename_mapping = {
    "LoadTransformationRules": "ImportExcelToRules",
    "InstancesFromRdfFileToSourceGraph": "ExtractGraphFromRdfFile",
    "InstancesFromRulesToSolutionGraph": "ExtractGraphFromRulesInstanceSheet",
    "GraphCapturingSheetToGraph": "ExtractGraphFromGraphCapturingSheet",
    "GenerateMockGraph": "ExtractGraphFromMockGraph",
    "InstancesFromJsonToGraph": "ExtractGraphFromJsonFile",
    "InstancesFromAvevaPiAF": "ExtractGraphFromAvevaPiAssetFramework",
    "DexpiToGraph": "ExtractGraphFromDexpiFile",
    "GenerateCDFAssetsFromGraph": "GenerateAssetsFromGraph",
    "GenerateCDFRelationshipsFromGraph": "GenerateRelationshipsFromGraph",
    "GenerateCDFNodesAndEdgesFromGraph": "GenerateNodesAndEdgesFromGraph",
    "UploadCDFAssets": "LoadAssetsToCDF",
    "UploadCDFRelationships": "LoadRelationshipsToCDF",
    "UploadCDFNodes": "LoadNodesToCDF",
    "UploadCDFEdges": "LoadEdgesToCDF",
    "CreateCDFLabels": "LoadLabelsToCDF",
    "OpenApiToRules": "ImportOpenApiToRules",
    "ArbitraryJsonYamlToRules": "ImportArbitraryJsonYamlToRules",
    "GraphToRules": "ImportGraphToRules",
    "OntologyToRules": "ImportOntologyToRules",
    "GraphQLSchemaFromRules": "ExportGraphQLSchemaFromRules",
    "OntologyFromRules": "ExportOntologyFromRules",
    "SHACLFromRules": "ExportSHACLFromRules",
    "GraphCaptureSpreadsheetFromRules": "ExportRulesToGraphCapturingSheet",
    "ExcelFromRules": "ExportRulesToExcel",
}


def rename_workflow_steps(workflow_path: Path, dry_run: bool) -> str:
    # Load the YAML file
    yaml = YAML()
    with workflow_path.open() as file:
        data = yaml.load(file)
    # Replace old step names with new ones
    for step in data["steps"]:
        if step["method"] in step_rename_mapping:
            print(f"Renaming step {step['method']} to {step_rename_mapping[step['method']]}")
            step["method"] = step_rename_mapping[step["method"]]

    # Save the updated YAML file
    with workflow_path.open("w") as file:
        yaml.dump(data, file)
    return "ok"


def migrate_all_workflow_names(workflows_path: str, dry_run: bool) -> str:
    # Get all workflow files
    workflow_files = Path(workflows_path).rglob("workflow.yaml")
    # Migrate each workflow file
    for workflow_file in workflow_files:
        print(f"Migrating {workflow_file}")
        rename_workflow_steps(workflow_file, dry_run)
    return "ok"


if __name__ == "__main__":
    # print current directory
    print("The current directory is", Path.cwd())
    migrate_all_workflow_names("data/workflows", True)



import shutil
from pathlib import Path

from cognite.neat.constants import EXAMPLE_GRAPHS, EXAMPLE_RULES, EXAMPLE_WORKFLOWS


def copy_examples_to_directory(target_data_dir: Path):
    """
    Copier over all the examples to the target_data_directory,
    without overwriting

    Args:
        target_data_dir : The target directory

    """

    print(f"Copying examples into {target_data_dir}")
    _copy_examples(EXAMPLE_RULES, target_data_dir / "rules")
    _copy_examples(EXAMPLE_GRAPHS, target_data_dir / "source-graphs")
    _copy_examples(EXAMPLE_WORKFLOWS, target_data_dir / "workflows")
    (target_data_dir / "staging").mkdir(exist_ok=True, parents=True)


def create_data_dir_structure(target_data_dir: Path):
    """
    Create the data directory structure in empty directory

    Args:
        target_data_dir : The target directory

    """

    (target_data_dir / "rules").mkdir(exist_ok=True, parents=True)
    (target_data_dir / "source-graphs").mkdir(exist_ok=True, parents=True)
    (target_data_dir / "staging").mkdir(exist_ok=True, parents=True)
    (target_data_dir / "workflows").mkdir(exist_ok=True, parents=True)


def _copy_examples(source_dir: Path, target_dir: Path):
    for current in source_dir.rglob("*"):
        if current.is_dir():
            continue
        relative = current.relative_to(source_dir)
        if not (target := target_dir / relative).exists():
            target.parent.mkdir(exist_ok=True, parents=True)
            shutil.copy2(current, target)



from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Generic, TypeVar

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    T_CogniteResourceList,
    T_WritableCogniteResource,
    T_WriteClass,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import (
    DataModelingId,
)
from cognite.client.data_classes.data_modeling.ids import (
    InstanceId,
    VersionedDataModelingId,
)
from cognite.client.utils.useful_types import SequenceNotStr

from .data_classes import RawTableID

T_ID = TypeVar("T_ID", bound=str | int | DataModelingId | InstanceId | VersionedDataModelingId | RawTableID)
T_WritableCogniteResourceList = TypeVar("T_WritableCogniteResourceList", bound=WriteableCogniteResourceList)


class ResourceLoader(
    ABC,
    Generic[T_ID, T_WriteClass, T_WritableCogniteResource, T_CogniteResourceList, T_WritableCogniteResourceList],
):
    resource_name: str

    def __init__(self, client: CogniteClient) -> None:
        self.client = client

    @classmethod
    @abstractmethod
    def get_id(cls, item: T_WriteClass | T_WritableCogniteResource) -> T_ID:
        raise NotImplementedError

    @classmethod
    def get_ids(cls, items: Sequence[T_WriteClass | T_WritableCogniteResource]) -> list[T_ID]:
        return [cls.get_id(item) for item in items]

    @abstractmethod
    def create(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def retrieve(self, ids: SequenceNotStr[T_ID]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def update(self, items: Sequence[T_WriteClass]) -> T_WritableCogniteResourceList:
        raise NotImplementedError

    @abstractmethod
    def delete(self, ids: SequenceNotStr[T_ID]) -> list[T_ID]:
        raise NotImplementedError

    def are_equal(self, local: T_WriteClass, remote: T_WritableCogniteResource) -> bool:
        return local == remote.as_write()



import logging
import shutil
from pathlib import Path
from typing import cast

from fastapi import APIRouter, Depends, HTTPException, Request, UploadFile
from fastapi.responses import FileResponse, JSONResponse

from cognite.neat.app.api.configuration import NEAT_APP
from cognite.neat.app.api.data_classes.rest import DownloadFromCdfRequest, RunWorkflowRequest, UploadToCdfRequest
from cognite.neat.workflows import WorkflowFullStateReport
from cognite.neat.workflows.base import WorkflowDefinition
from cognite.neat.workflows.migration.wf_manifests import migrate_wf_manifest
from cognite.neat.workflows.model import FlowMessage
from cognite.neat.workflows.steps.data_contracts import SolutionGraph, SourceGraph
from cognite.neat.workflows.steps.step_model import DataContract
from cognite.neat.workflows.utils import get_file_hash

router = APIRouter()


@router.post("/api/workflow/start")
def start_workflow(request: RunWorkflowRequest):
    logging.info("Starting workflow endpoint")
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    start_status = NEAT_APP.workflow_manager.start_workflow_instance(
        request.name, sync=request.sync, flow_msg=FlowMessage()
    )
    result = {"workflow_instance": None, "is_success": start_status.is_success, "status_text": start_status.status_text}
    return {"result": result}


@router.get("/api/workflow/stats/{workflow_name}", response_model=WorkflowFullStateReport)
def get_workflow_stats(
    workflow_name: str,
) -> WorkflowFullStateReport | None | dict[str, str]:
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        raise HTTPException(status_code=404, detail="workflow not found")
    return workflow.get_state()


@router.get("/api/workflow/workflows")
def get_workflows():
    return {"workflows": NEAT_APP.workflow_manager.get_list_of_workflows()}


@router.get("/api/workflow/files/{workflow_name}")
def get_workflow_files(workflow_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        raise HTTPException(status_code=404, detail="workflow not found")
    return {"files": workflow.get_list_of_workflow_artifacts()}


@router.post("/api/workflow/package/{workflow_name}")
def package_workflow(workflow_name: str):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    package_file = NEAT_APP.cdf_store.package_workflow(workflow_name)
    hash = get_file_hash(NEAT_APP.config.data_store_path / "workflows" / package_file)
    return {"package": package_file, "hash": hash}


@router.post("/api/workflow/context-cleanup/{workflow_name}")
def cleanup_workflow_data(workflow_name: str):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is not None:
        workflow.cleanup_workflow_context()
    return {"result": "ok"}


@router.post("/api/workflow/create")
def create_new_workflow(request: WorkflowDefinition):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    new_workflow_definition = NEAT_APP.workflow_manager.create_new_workflow(
        request.name, request.description, "manifest"
    )
    return {"workflow": new_workflow_definition}


@router.delete("/api/workflow/{workflow_name}")
def delete_workflow(workflow_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    NEAT_APP.workflow_manager.delete_workflow(workflow_name)
    return {"result": "ok"}


@router.get("/api/workflow/executions")
def get_list_of_workflow_executions():
    return {"executions": NEAT_APP.cdf_store.get_list_of_workflow_executions_from_cdf()}


@router.get("/api/workflow/detailed-execution-report/{execution_id}")
def get_detailed_execution(execution_id: str):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    return {"report": NEAT_APP.cdf_store.get_detailed_workflow_execution_report_from_cdf(execution_id)}


@router.post("/api/workflow/reload-workflows")
def reload_workflows():
    NEAT_APP.workflow_manager.load_workflows_from_storage()
    NEAT_APP.triggers_manager.reload_all_triggers()
    return {"result": "ok", "workflows": NEAT_APP.workflow_manager.get_list_of_workflows()}


@router.post("/api/workflow/reload-single-workflow/{workflow_name}")
def reload_single_workflows(workflow_name: str):
    NEAT_APP.workflow_manager.load_single_workflow_from_storage(workflow_name)
    NEAT_APP.triggers_manager.reload_all_triggers()
    return {"result": "ok", "workflows": NEAT_APP.workflow_manager.get_list_of_workflows()}


@router.get("/api/workflow/workflow-definition/{workflow_name}")
def get_workflow_definition(workflow_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"error": "Workflow is not initialized"}
    return {"definition": workflow.get_workflow_definition()}


@router.get("/api/workflow/workflow-src/{workflow_name}/{file_name}")
def get_workflow_src(workflow_name: str, file_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    # Todo: Thi sis a bug in the API. The method below does not exist
    src = NEAT_APP.workflow_manager.get_workflow_src(workflow_name, file_name=file_name)  # type: ignore[attr-defined]
    return FileResponse(src, media_type="text/plain")


@router.post("/api/workflow/workflow-definition/{workflow_name}")
def update_workflow_definition(workflow_name: str, request: WorkflowDefinition):
    if NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    wf = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if wf is not None:
        wf.cleanup_workflow_context()
    NEAT_APP.workflow_manager.update_workflow(workflow_name, request)
    NEAT_APP.workflow_manager.save_workflow_to_storage(workflow_name)
    return {"result": "ok"}


@router.post("/api/workflow/upload-wf-to-cdf/{workflow_name}")
def upload_workflow_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    NEAT_APP.cdf_store.save_workflow_to_cdf(
        workflow_name, changed_by=request.author, comments=request.comments, tag=request.tag
    )
    return {"result": "ok"}


@router.post("/api/workflow/upload-rules-cdf/{workflow_name}")
def upload_rules_to_cdf(workflow_name: str, request: UploadToCdfRequest):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    file_path = Path(NEAT_APP.config.rules_store_path, request.file_name)
    NEAT_APP.cdf_store.save_resource_to_cdf(
        workflow_name, "neat-wf-rules", file_path, changed_by=request.author, comments=request.comments
    )
    return {"result": "ok"}


@router.post("/api/workflow/download-wf-from-cdf")
def download_wf_from_cdf(request: DownloadFromCdfRequest):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    NEAT_APP.cdf_store.load_workflows_from_cdf(request.file_name, request.version)
    return {"result": "ok"}


@router.post("/api/workflow/download-rules-from-cdf")
def download_rules_to_cdf(request: DownloadFromCdfRequest):
    if NEAT_APP.cdf_store is None:
        return {"error": "NeatApp is not initialized"}
    NEAT_APP.cdf_store.load_rules_file_from_cdf(request.file_name, request.version)
    return {"file_name": request.file_name, "hash": request.version}


@router.post("/api/workflow/migrate-workflow")
def migrate_workflow():
    return migrate_wf_manifest(NEAT_APP.config.data_store_path)


@router.get("/api/workflow/pre-cdf-assets/{workflow_name}")
def get_pre_cdf_assets(workflow_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "Workflow Manager is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"assets": []}
    return {"assets": workflow.data["CategorizedAssets"]}


@router.get("/api/workflow/context/{workflow_name}")
def get_context(workflow_name: str):
    if NEAT_APP.workflow_manager is None:
        return {"error": "Workflow Manager is not initialized"}
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"error": "Workflow is not initialized"}
    context = workflow.get_context()
    objects_in_context = []
    for key, value in context.items():
        objects_in_context.append({"name": key, "type": type(value).__name__})
    return {"context": objects_in_context}


@router.get("/api/workflow/context/{workflow_name}/object_name/{object_name}")
def get_context_object(workflow_name: str, object_name: str):
    """Get context item from workflow. Should be used for debugging and troubleshooting only."""
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return
    context = workflow.get_context()
    if object_name not in context:
        return {"error": f"Item {object_name} is not found in workflow context"}

    if object_name == "SourceGraph" or object_name == "SolutionGraph":
        return {"object": cast(SourceGraph | SolutionGraph, context[object_name]).graph.diagnostic_report()}

    cobject = context[object_name]
    if isinstance(cobject, DataContract):
        return {"object": cobject.model_dump()}
    else:
        return {"object": cobject}


@router.get("/api/workflow/registered-steps")
def get_steps():
    steps_registry = NEAT_APP.workflow_manager.get_steps_registry()
    return {"steps": steps_registry.get_list_of_steps()}


@router.post("/api/workflow/file/{workflow_name}")
async def upload_file(file: UploadFile, workflow_name: str):
    if NEAT_APP.workflow_manager is None or NEAT_APP.workflow_manager.data_store_path is None:
        return JSONResponse(content={"error": "Workflow Manager is not initialized"}, status_code=400)
    try:
        upload_dir = NEAT_APP.workflow_manager.data_store_path / "workflows" / workflow_name
        # Create a directory to store uploaded files if it doesn't exist

        # Define the file path where the uploaded file will be saved
        if file.filename is None:
            return JSONResponse(content={"message": "File name is not provided"}, status_code=400)
        file_path = upload_dir / file.filename

        # Save the uploaded file to the specified path
        with file_path.open("wb") as buffer:
            shutil.copyfileobj(file.file, buffer)

        if file.filename.endswith(".py"):
            NEAT_APP.workflow_manager.steps_registry.load_workflow_step_classes(workflow_name)

        return JSONResponse(content={"message": "File uploaded successfully"}, status_code=200)
    except Exception as e:
        return JSONResponse(content={"message": f"An error occurred: {e!s}"}, status_code=500)


async def get_body(request: Request):
    return await request.body()


fast_api_depends = Depends(get_body)


@router.post("/api/workflow/{workflow_name}/http_trigger/{step_id}")
def http_trigger_start_workflow(workflow_name: str, step_id: str, request: Request, body: bytes = fast_api_depends):
    if NEAT_APP.triggers_manager is None:
        return JSONResponse(content={"error": "Triggers Manager is not initialized"}, status_code=400)
    return NEAT_APP.triggers_manager.start_workflow_from_http_request(workflow_name, step_id, request, body)


@router.post("/api/workflow/{workflow_name}/resume/{step_id}/{instance_id}")
def http_trigger_resume_workflow(
    workflow_name: str, step_id: str, instance_id: str, request: Request, body: bytes = fast_api_depends
):
    if NEAT_APP.triggers_manager is None:
        return JSONResponse(content={"error": "Triggers Manager is not initialized"}, status_code=400)
    return NEAT_APP.triggers_manager.resume_workflow_from_http_request(
        workflow_name, step_id, instance_id, request, body
    )



from pydantic import BaseModel, field_validator


class CogniteClientConfig(BaseModel):
    project: str = "dev"
    client_id: str = "neat"
    base_url: str = "https://api.cognitedata.com"
    scopes: list[str] = ["project:read", "project:write"]
    timeout: int = 60
    max_workers: int = 3

    @field_validator("scopes", mode="before")
    def string_to_list(cls, value):
        return [value] if isinstance(value, str) else value


class InteractiveCogniteClient(CogniteClientConfig):
    authority_url: str
    redirect_port: int = 53_000


class ServiceCogniteClient(CogniteClientConfig):
    token_url: str = "https://login.microsoftonline.com/common/oauth2/token"
    client_secret: str = "secret"



from .base import RoleTypes
from .dms_architect_rules import DMSRules
from .dms_schema import DMSSchema
from .domain_rules import DomainRules
from .information_rules import InformationRules

RULES_PER_ROLE: dict[RoleTypes, type[DomainRules] | type[InformationRules] | type[DMSRules]] = {
    RoleTypes.domain_expert: DomainRules,
    RoleTypes.information_architect: InformationRules,
    RoleTypes.dms_architect: DMSRules,
}


__all__ = ["DomainRules", "InformationRules", "DMSRules", "RULES_PER_ROLE", "DMSSchema"]



import sys
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models._rules._types import ClassEntity

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

T_Rules = TypeVar("T_Rules", bound=Rules)


class BaseAnalysis(ABC, Generic[T_Rules]):
    @abstractmethod
    def subset_rules(self, desired_classes: set[ClassEntity]) -> T_Rules:
        raise NotImplementedError()


class DataModelingScenario(StrEnum):
    from_scratch = "from scratch"
    build_solution = "build solution"
    extend_reference = "extend reference"



import logging
from collections.abc import Iterable
from itertools import islice
from typing import Literal, cast, overload

from cognite.client import CogniteClient
from cognite.client.data_classes.data_modeling import EdgeApply, InstanceApply, NodeApply
from pydantic_core import ErrorDetails

from cognite.neat.exceptions import NeatException
from cognite.neat.graph.stores import NeatGraphStoreBase
from cognite.neat.graph.transformation.query_generator.sparql import triples2dictionary
from cognite.neat.rules.exporter._rules2dms import DMSSchemaComponents
from cognite.neat.rules.exporter._rules2pydantic_models import add_class_prefix_to_xid, rules_to_pydantic_models
from cognite.neat.rules.models.rules import Rules
from cognite.neat.utils.utils import chunker, datetime_utc_now, retry_decorator

from ._base import CogniteLoader


class DMSLoader(CogniteLoader[InstanceApply]):
    """Loads a Neat Graph into CDF as nodes and edges.

    Args:
        rules: Rules object
        graph_store: Graph store
        add_class_prefix: Add class prefix to external_id. Defaults to False.

    """

    def __init__(self, rules: Rules, graph_store: NeatGraphStoreBase, add_class_prefix: bool = False):
        super().__init__(rules, graph_store)
        self.add_class_prefix = add_class_prefix

    @overload
    def load(self, stop_on_exception: Literal[True]) -> Iterable[InstanceApply]:
        ...

    @overload
    def load(self, stop_on_exception: Literal[False] = False) -> Iterable[InstanceApply | ErrorDetails]:
        ...

    def load(self, stop_on_exception: bool = False) -> Iterable[InstanceApply | ErrorDetails]:
        """Load the graph with data."""
        if self.rules.metadata.namespace is None:
            raise ValueError("Namespace is not defined in transformation rules metadata")

        data_model = DMSSchemaComponents.from_rules(self.rules)
        pydantic_models = rules_to_pydantic_models(self.rules)

        exclude = {
            class_name
            for class_name in self.rules.classes
            if f"{self.rules.space}:{class_name}" not in data_model.containers
        }

        for class_name, triples in self._iterate_class_triples(exclude_classes=exclude):
            logging.info(f"<DMSLoader> Processing class : {class_name}")
            counter = 0
            start_time = datetime_utc_now()
            for instance_dict in triples2dictionary(triples).values():
                counter += 1
                try:
                    instance = pydantic_models[class_name].from_dict(instance_dict)  # type: ignore[attr-defined]
                    if self.add_class_prefix:
                        instance.external_id = add_class_prefix_to_xid(
                            class_name=type(instance).__name__, external_id=instance.external_id
                        )
                    new_node = instance.to_node(data_model, self.add_class_prefix)  # type: ignore[attr-defined]
                    is_valid, reason = is_node_valid(new_node)
                    if is_valid:
                        yield new_node
                    else:
                        yield ErrorDetails(
                            input=instance_dict["external_id"],
                            loc=tuple(["Nodes"]),
                            msg=f"Not valid node {new_node.external_id}. Reason: {reason}",
                            type="Node validation error",
                        )
                        continue

                    new_edges = instance.to_edge(data_model, self.add_class_prefix)
                    for new_edge in new_edges:
                        is_valid, reason = is_edge_valid(new_edge)
                        if is_valid:
                            yield new_edge
                        else:
                            yield ErrorDetails(
                                input=instance_dict["external_id"],
                                loc=tuple(["Edges"]),
                                msg=f"Not valid edge {new_edge.external_id}. Reason: {reason}",
                                type="Edge validation error",
                            )
                            continue

                    delta_time = datetime_utc_now() - start_time
                    delta_time = (delta_time.seconds * 1000000 + delta_time.microseconds) / 1000

                except Exception as e:
                    logging.error(
                        f"Instance {instance_dict['external_id']} of {class_name}"
                        f" cannot be resolved to nodes and edges. Reason: {e}"
                    )
                    if stop_on_exception:
                        raise e

                    if isinstance(e, NeatException):
                        yield e.to_error_dict()
                    else:
                        yield ErrorDetails(
                            input=instance_dict["external_id"],
                            loc=tuple(["rdf2nodes_and_edges"]),
                            msg=str(e),
                            type=f"Exception of type {type(e).__name__} occurred  \
                                        when processing instance of {class_name}",
                        )

    def load_to_cdf(
        self, client: CogniteClient, batch_size: int | None = 1000, max_retries: int = 1, retry_delay: int = 3
    ) -> None:
        """Uploads nodes to CDF

        Args:
            client: Instance of CogniteClient
            batch_size: Size of batch. Default to 1000.
            max_retries: Maximum times to retry the upload. Default to 1.
            retry_delay: Time delay before retrying the upload. Default to 3.

        !!! note "batch_size"
            If batch size is set to 1 or None, all nodes will be pushed to CDF in one go.
        """
        if batch_size is None:
            logging.info("Batch size not set, pushing all nodes and edges to CDF in one go!")
            nodes, edges, errors = self.as_nodes_and_edges(stop_on_exception=False)

            @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-instances")
            def create_instances():
                client.data_modeling.instances.apply(
                    nodes=nodes, edges=edges, auto_create_start_nodes=True, auto_create_end_nodes=True
                )

            create_instances()
            return
        logging.info(f"Uploading nodes in batches of {batch_size}")
        for instances in _batched(self.load(stop_on_exception=False), batch_size):
            nodes = [instance for instance in instances if isinstance(instance, NodeApply)]
            edges = [instance for instance in instances if isinstance(instance, EdgeApply)]
            # Todo make _micro_batch_push handle both nodes and edges simultaneously
            _micro_batch_push(
                client, nodes, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay
            )
            _micro_batch_push(
                client, edges, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay
            )

    def as_nodes_and_edges(
        self, stop_on_exception: bool = False
    ) -> tuple[list[NodeApply], list[EdgeApply], list[ErrorDetails]]:
        nodes = []
        edges = []
        exceptions: list[ErrorDetails] = []
        for instance in self.load(stop_on_exception):  # type: ignore[call-overload]
            if isinstance(instance, NodeApply):
                nodes.append(instance)
            elif isinstance(instance, EdgeApply):
                edges.append(instance)
            elif isinstance(instance, dict):
                exceptions.append(cast(ErrorDetails, instance))
            else:
                raise ValueError(f"Unknown instance type: {type(instance)}")
        return nodes, edges, exceptions


def _batched(iterable: Iterable, size: int):
    "Batch data into lists of length n. The last batch may be shorter."
    # batched('ABCDEFG', 3) --> ABC DEF G
    it = iter(iterable)
    while True:
        batch = list(islice(it, size))
        if not batch:
            return
        yield batch


def is_node_valid(node: NodeApply) -> tuple[bool, str]:
    return is_valid_external_id(node.external_id)


def is_edge_valid(edge: EdgeApply) -> tuple[bool, str]:
    for external_id in [edge.external_id, edge.start_node.external_id, edge.end_node.external_id]:
        is_valid, reason = is_valid_external_id(external_id)
        if not is_valid:
            return False, reason
    return True, ""


def is_valid_external_id(external_id: str) -> tuple[bool, str]:
    if external_id is None or external_id == "" or len(external_id) >= 255:
        return False, f"external_id {external_id} is empty of too long"
    return True, ""


def upload_nodes(
    client: CogniteClient, nodes: list[NodeApply], batch_size: int = 1000, max_retries: int = 1, retry_delay: int = 3
):
    """Uploads nodes to CDF

    Args:
        client: Instance of CogniteClient
        nodes: List of nodes to upload to CDF
        batch_size: Size of batch. Defaults to 1000.
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.

    !!! note "batch_size"
        If batch size is set to 1 or None, all nodes will be pushed to CDF in one go.
    """
    if batch_size:
        logging.info(f"Uploading nodes in batches of {batch_size}")
        _micro_batch_push(client, nodes, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay)

    else:
        logging.info("Batch size not set, pushing all nodes to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-nodes")
        def create_nodes():
            client.data_modeling.instances.apply(nodes=nodes)

        create_nodes()


def upload_edges(
    client: CogniteClient, edges: list[EdgeApply], batch_size: int = 5000, max_retries: int = 1, retry_delay: int = 3
):
    """Uploads edges to CDF

    Args:
        client: Instance of CogniteClient
        edges: List of edges to upload to CDF
        batch_size: Size of batch. Defaults to 5000.
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.

    !!! note "batch_size"
        If batch size is set to 1 or None, all edges will be pushed to CDF in one go.

    """
    if batch_size:
        logging.info(f"Uploading edges in batches of {batch_size}")
        _micro_batch_push(client, edges, batch_size, message="Upload", max_retries=max_retries, retry_delay=retry_delay)

    else:
        logging.info("Batch size not set, pushing all edges to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-edges")
        def create_nodes():
            client.data_modeling.instances.apply(edges=edges, auto_create_start_nodes=True, auto_create_end_nodes=True)

        create_nodes()


def _micro_batch_push(
    client: CogniteClient,
    nodes_or_edges: list[NodeApply] | list[EdgeApply],
    batch_size: int = 1000,
    message: str = "Upload",
    max_retries: int = 1,
    retry_delay: int = 3,
):
    """Uploads nodes or edges in batches

    Args:
        client: Instance of CogniteClient
        nodes_or_edges: List of nodes or edges
        batch_size: Size of batch. Defaults to 1000.
        message: Message to logged. Defaults to "Upload".
        max_retries: Maximum times to retry the upload. Defaults to 1.
        retry_delay: Time delay before retrying the upload. Defaults to 3.
    """
    total = len(nodes_or_edges)
    counter = 0

    if nodes_or_edges and isinstance(nodes_or_edges[0], NodeApply):
        push_type = "nodes"
    elif nodes_or_edges and isinstance(nodes_or_edges[0], EdgeApply):
        push_type = "edges"
    else:
        raise ValueError("nodes_or_edges must be a list of NodeApply or EdgeApply objects")

    for batch in chunker(nodes_or_edges, batch_size):
        counter += len(batch)
        start_time = datetime_utc_now()

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name=f"microbatch-{push_type}")
        def upsert_nodes_or_edges(upload_batch):
            if push_type == "nodes":
                client.data_modeling.instances.apply(nodes=upload_batch)
            elif push_type == "edges":
                client.data_modeling.instances.apply(
                    edges=upload_batch, auto_create_start_nodes=True, auto_create_end_nodes=True
                )

        upsert_nodes_or_edges(batch)

        delta_time = (datetime_utc_now() - start_time).seconds

        msg = f"{message} {counter} of {total} {push_type}, batch processing time: {delta_time:.2f} "
        msg += f"seconds ETC: {delta_time * (total - counter) / (60*batch_size) :.2f} minutes"
        logging.info(msg)



from collections.abc import Iterable, Sequence
from itertools import groupby
from typing import cast, overload

from cognite.client.data_classes import (
    Database,
    DatabaseList,
    DatabaseWrite,
    DatabaseWriteList,
    Transformation,
    TransformationList,
    TransformationWrite,
    TransformationWriteList,
)
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils.useful_types import SequenceNotStr

from ._base import ResourceLoader
from .data_classes import RawTable, RawTableID, RawTableList, RawTableWrite, RawTableWriteList


class TransformationLoader(
    ResourceLoader[str, TransformationWrite, Transformation, TransformationWriteList, TransformationList]
):
    resource_name = "transformations"

    @classmethod
    def get_id(cls, item: Transformation | TransformationWrite) -> str:
        if item.external_id is None:
            raise ValueError(f"Transformation {item} does not have an external_id")
        return item.external_id

    def create(self, items: Sequence[TransformationWrite]) -> TransformationList:
        return self.client.transformations.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> TransformationList:
        return self.client.transformations.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: Sequence[TransformationWrite]) -> TransformationList:
        return self.client.transformations.update(items)

    def delete(self, ids: SequenceNotStr[str]) -> list[str]:
        existing = self.retrieve(ids)
        self.client.transformations.delete(external_id=ids, ignore_unknown_ids=True)
        return existing.as_external_ids()


class RawDatabaseLoader(ResourceLoader[str, DatabaseWrite, Database, DatabaseWriteList, DatabaseList]):
    resource_name = "databases"

    @classmethod
    def get_id(cls, item: Database | DatabaseWrite) -> str:
        if item.name is None:
            raise ValueError(f"Database {item} does not have a name")
        return item.name

    def create(self, items: Sequence[DatabaseWrite]) -> DatabaseList:
        return self.client.raw.databases.create([item.name for item in items if item.name is not None])

    def retrieve(self, ids: SequenceNotStr[str]) -> DatabaseList:
        all_databases = self.client.raw.databases.list(limit=-1)
        return DatabaseList([db for db in all_databases if db.name in ids])

    def update(self, items: Sequence[DatabaseWrite]) -> DatabaseList:
        if not items:
            return DatabaseList([])
        raise NotImplementedError("The CDF API does not support updating a RAW database.")

    def delete(self, ids: SequenceNotStr[str]) -> list[str]:
        existing_databases = self.retrieve(ids)
        existing_names = {item.name for item in existing_databases}
        self.client.raw.databases.delete([name for name in ids if name in existing_names])
        return existing_databases.as_names()


class RawTableLoader(ResourceLoader[RawTableID, RawTableWrite, RawTable, RawTableWriteList, RawTableList]):
    resource_name = "tables"

    @classmethod
    def get_id(cls, item: RawTable | RawTableWrite) -> RawTableID:
        return item.as_id()

    @overload
    def _groupby_database(self, items: Sequence[RawTableWrite]) -> Iterable[tuple[str, Iterable[RawTableWrite]]]:
        ...

    @overload
    def _groupby_database(self, items: SequenceNotStr[RawTableID]) -> Iterable[tuple[str, Iterable[RawTableID]]]:
        ...

    def _groupby_database(
        self, items: Sequence[RawTableWrite] | SequenceNotStr[RawTableID]
    ) -> Iterable[tuple[str, Iterable[RawTableWrite] | Iterable[RawTableID]]]:
        return cast(
            Iterable[tuple[str, Iterable[RawTableID] | Iterable[RawTableWrite]]],
            groupby(sorted(items, key=lambda x: x.database or ""), lambda x: x.database or ""),
        )

    def create(self, items: Sequence[RawTableWrite]) -> RawTableList:
        existing = set(self.retrieve([table.as_id() for table in items]).as_ids())
        output = RawTableList([])
        for db_name, tables in self._groupby_database(items):
            to_create = [table.name for table in tables if table.name if table.as_id() not in existing]
            if not to_create:
                continue
            created = self.client.raw.tables.create(db_name=db_name, name=to_create)
            for table in created:
                output.append(
                    RawTable(
                        name=table.name, database=db_name, created_time=table.created_time, cognite_client=self.client
                    )
                )
        return output

    def retrieve(self, ids: SequenceNotStr[RawTableID]) -> RawTableList:
        output = RawTableList([])
        for db_name, id_group in self._groupby_database(ids):
            try:
                all_tables = self.client.raw.tables.list(db_name, limit=-1)
            except CogniteAPIError as e:
                if e.code == 404 and e.message.startswith("Following databases not found"):
                    continue
            looking_for = {table_id.table for table_id in id_group if table_id.table is not None}
            output.extend(
                [
                    RawTable(
                        name=table.name, database=db_name, created_time=table.created_time, cognite_client=self.client
                    )
                    for table in all_tables
                    if table.name in looking_for
                ]
            )
        return output

    def update(self, items: Sequence[RawTableWrite]) -> RawTableList:
        if not items:
            return RawTableList([])
        raise NotImplementedError("The CDF API does not support updating a RAW table.")

    def delete(self, ids: SequenceNotStr[RawTableID]) -> list[RawTableID]:
        existing_tables = self.retrieve(ids)
        existing_names = {item.name for item in existing_tables}
        for db_name, id_group in self._groupby_database(ids):
            self.client.raw.tables.delete(
                db_name=db_name,
                name=[
                    table_id.table
                    for table_id in id_group
                    if table_id.table is not None and table_id.table in existing_names
                ],
            )
        return existing_tables.as_ids()



from collections import Counter
from collections.abc import Callable, Sequence

import cognite.neat.rules.issues.importing
from cognite.neat.rules import issues
from cognite.neat.rules.importers._dtdl2rules.spec import (
    DTMI,
    Command,
    CommandV2,
    Component,
    DTDLBase,
    Enum,
    Interface,
    Object,
    Property,
    PropertyV2,
    Relationship,
    Schema,
    Telemetry,
    TelemetryV2,
)
from cognite.neat.rules.issues import IssueList, ValidationIssue
from cognite.neat.rules.models._rules._types import (
    XSD_VALUE_TYPE_MAPPINGS,
    ClassEntity,
    ParentClassEntity,
    XSDValueType,
)
from cognite.neat.rules.models._rules.information_rules import InformationClass, InformationProperty


class _DTDLConverter:
    def __init__(self, issues: list[ValidationIssue] | None = None) -> None:
        self.issues: IssueList = IssueList(issues or [])
        self.properties: list[InformationProperty] = []
        self.classes: list[InformationClass] = []
        self._item_by_id: dict[DTMI, DTDLBase] = {}

        self._method_by_type: dict[type[DTDLBase], Callable[[DTDLBase, str | None], None]] = {
            Interface: self.convert_interface,  # type: ignore[dict-item]
            Property: self.convert_property,  # type: ignore[dict-item]
            PropertyV2: self.convert_property,  # type: ignore[dict-item]
            Relationship: self.convert_relationship,  # type: ignore[dict-item]
            Object: self.convert_object,  # type: ignore[dict-item]
            Telemetry: self.convert_telemetry,  # type: ignore[dict-item]
            TelemetryV2: self.convert_telemetry,  # type: ignore[dict-item]
            Command: self.convert_command,  # type: ignore[dict-item]
            CommandV2: self.convert_command,  # type: ignore[dict-item]
            Component: self.convert_component,  # type: ignore[dict-item]
        }

    def get_most_common_prefix(self) -> str:
        if not self.classes:
            raise ValueError("No classes found")
        counted = Counter(cls_.class_.prefix for cls_ in self.classes if isinstance(cls_.class_.prefix, str))
        if not counted:
            raise ValueError("No prefixes found")
        return counted.most_common(1)[0][0]

    def convert(self, items: Sequence[DTDLBase]) -> None:
        self._item_by_id.update({item.id_: item for item in items if item.id_ is not None})
        # Update schema objects which are reusable
        self._item_by_id.update(
            {
                schema.id_: schema
                for item in items
                if isinstance(item, Interface)
                for schema in item.schemas or []
                if isinstance(schema.id_, DTMI) and isinstance(schema, DTDLBase)
            }
        )

        for item in items:
            self.convert_item(item)

    def convert_item(self, item: DTDLBase, parent: str | None = None) -> None:
        convert_method = self._method_by_type.get(type(item))
        if convert_method is not None:
            convert_method(item, parent)
        else:
            self.issues.append(
                issues.importing.UnknownComponentWarning(
                    component_type=item.type,
                    instance_name=item.display_name,
                    instance_id=item.id_.model_dump() if item.id_ else None,
                )
            )

    def convert_interface(self, item: Interface, _: str | None) -> None:
        class_ = InformationClass(
            class_=item.id_.as_class_id(),
            name=item.display_name,
            description=item.description,
            comment=item.comment,
            parent=[ParentClassEntity.from_raw(parent.as_class_id()) for parent in item.extends or []] or None,
        )
        self.classes.append(class_)
        for sub_item_or_id in item.contents or []:
            if isinstance(sub_item_or_id, DTMI) and sub_item_or_id not in self._item_by_id:
                self.issues.append(
                    issues.importing.UnknownPropertyWarning(
                        component_type=item.type,
                        property_name=sub_item_or_id.path[-1],
                        instance_name=item.display_name,
                        instance_id=item.id_.model_dump(),
                    )
                )
            elif isinstance(sub_item_or_id, DTMI):
                sub_item = self._item_by_id[sub_item_or_id]
                self.convert_item(sub_item, class_.class_.versioned_id)
            else:
                self.convert_item(sub_item_or_id, class_.class_.versioned_id)
        # interface.schema objects are handled in the convert method

    def convert_property(
        self, item: Property | Telemetry, parent: str | None, min_count: int | None = 0, max_count: int | None = 1
    ) -> None:
        if parent is None:
            self._missing_parent_warning(item)
            return None
        value_type = self.schema_to_value_type(item.schema_, item)
        if value_type is None:
            return None

        prop = InformationProperty(
            class_=ClassEntity.from_raw(parent),
            property_=item.name,
            name=item.display_name,
            description=item.description,
            comment=item.comment,
            value_type=value_type,
            min_count=min_count,
            max_count=max_count,
        )
        self.properties.append(prop)

    def _missing_parent_warning(self, item):
        self.issues.append(
            cognite.neat.rules.issues.importing.MissingParentDefinitionError(
                component_type=item.type,
                instance_name=item.display_name,
                instance_id=item.id_.model_dump() if item.id_ else None,
            )
        )

    def convert_telemetry(self, item: Telemetry, parent: str | None) -> None:
        return self.convert_property(
            item,
            parent,
        )

    def convert_command(self, item: Command | CommandV2, parent: str | None) -> None:
        if parent is None:
            self._missing_parent_warning(item)
            return None
        if item.request is None:
            self.issues.append(
                issues.importing.UnknownSubComponentWarning(
                    component_type=item.type,
                    sub_component="request",
                    instance_name=item.display_name,
                    instance_id=item.id_.model_dump() if item.id_ else None,
                )
            )
            return None
        if item.response is not None:
            # Currently, we do not know how to handle response
            self.issues.append(
                issues.importing.IgnoredComponentWarning(
                    identifier=f"{parent}.response",
                    reason="Neat does not have a concept of response for commands. This will be ignored.",
                )
            )
        value_type = self.schema_to_value_type(item.request.schema_, item)
        if value_type is None:
            return
        prop = InformationProperty(
            class_=ClassEntity.from_raw(parent),
            property_=item.name,
            name=item.display_name,
            description=item.description,
            comment=item.comment,
            value_type=value_type,
            min_count=0,
            max_count=1,
        )
        self.properties.append(prop)

    def convert_component(self, item: Component, parent: str | None) -> None:
        if parent is None:
            self._missing_parent_warning(item)
            return None

        value_type = self.schema_to_value_type(item.schema_, item)
        if value_type is None:
            return
        prop = InformationProperty(
            class_=ClassEntity.from_raw(parent),
            property_=item.name,
            name=item.display_name,
            description=item.description,
            comment=item.comment,
            value_type=value_type,
            min_count=0,
            max_count=1,
        )
        self.properties.append(prop)

    def convert_relationship(self, item: Relationship, parent: str | None) -> None:
        if parent is None:
            self._missing_parent_warning(item)
            return None
        if item.target is not None:
            value_type: XSDValueType | ClassEntity
            if item.target in self._item_by_id:
                value_type = item.target.as_class_id()
            else:
                # Falling back to json
                self.issues.append(
                    cognite.neat.rules.issues.importing.MissingIdentifierError(
                        component_type="Unknown",
                        instance_name=item.target.model_dump(),
                        instance_id=item.target.model_dump(),
                    )
                )
                value_type = XSD_VALUE_TYPE_MAPPINGS["json"]

            prop = InformationProperty(
                class_=ClassEntity.from_raw(parent),
                property_=item.name,
                name=item.display_name,
                description=item.description,
                min_count=item.min_multiplicity or 0,
                max_count=item.max_multiplicity or 1,
                comment=item.comment,
                value_type=value_type,
            )
            self.properties.append(prop)
        elif item.properties is not None:
            for prop_ in item.properties or []:
                self.convert_property(prop_, parent, item.min_multiplicity, item.max_multiplicity)

    def convert_object(self, item: Object, _: str | None) -> None:
        if item.id_ is None:
            self.issues.append(
                cognite.neat.rules.issues.importing.MissingIdentifierError(
                    component_type=item.type,
                    instance_name=item.display_name,
                )
            )
            return None

        class_ = InformationClass(
            class_=item.id_.as_class_id(),
            name=item.display_name,
            description=item.description,
            comment=item.comment,
        )
        self.classes.append(class_)

        for field_ in item.fields or []:
            value_type = self.schema_to_value_type(field_.schema_, item)
            if value_type is None:
                continue
            prop = InformationProperty(
                class_=class_.class_,
                name=field_.name,
                description=field_.description,
                property_=field_.name,
                value_type=value_type,
                min_count=0,
                max_count=1,
            )
            self.properties.append(prop)

    def schema_to_value_type(
        self, schema: Schema | Interface | DTMI | None, item: DTDLBase
    ) -> XSDValueType | ClassEntity | None:
        input_type = self._item_by_id.get(schema) if isinstance(schema, DTMI) else schema

        if isinstance(input_type, Enum):
            return XSD_VALUE_TYPE_MAPPINGS["string"]
        elif isinstance(input_type, str) and input_type in XSD_VALUE_TYPE_MAPPINGS:
            return XSD_VALUE_TYPE_MAPPINGS[input_type]
        elif isinstance(input_type, str):
            self.issues.append(
                cognite.neat.rules.issues.importing.UnsupportedPropertyTypeError(
                    component_type=item.type,
                    property_type=input_type,
                    property_name="schema",
                    instance_name=item.display_name,
                    instance_id=item.id_.model_dump() if item.id_ else None,
                )
            )
            return None
        elif isinstance(input_type, Object | Interface):
            if input_type.id_ is None:
                self.issues.append(
                    cognite.neat.rules.issues.importing.MissingIdentifierError(
                        component_type=input_type.type,
                        instance_name=input_type.display_name,
                    )
                )
                return XSD_VALUE_TYPE_MAPPINGS["json"]
            else:
                if isinstance(input_type, Object):
                    self.convert_object(input_type, None)
                return ClassEntity.from_raw(input_type.id_.as_class_id())
        else:
            self.issues.append(
                issues.importing.UnknownPropertyWarning(
                    component_type=item.type,
                    property_name="schema",
                    instance_name=item.display_name,
                    instance_id=item.id_.model_dump() if item.id_ else None,
                )
            )
            return None
